"""
Fetch API calls wrapping Tradier

Supported environment variables:

::

    # verbose logging in this module
    export DEBUG_FETCH=1

"""

import json
import datetime
import requests
import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.options_dates as opt_dates
import analysis_engine.url_helper as url_helper
import analysis_engine.dataset_scrub_utils as scrub_utils
import analysis_engine.td.consts as td_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def fetch_calls(
        work_dict=None,
        ticker=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_calls

    Fetch Tradier option calls for a ticker and
    return a tuple: (status, ``pandas.DataFrame``)

    .. code-block:: python

        import analysis_engine.td.fetch_api as td_fetch

        # Please set the TD_TOKEN environment variable to your token
        calls_status, calls_df = td_fetch.fetch_calls(
            ticker='SPY')

        print(f'Fetched SPY Option Calls from Tradier status={calls_status}:')
        print(calls_df)

    :param work_dict: optional - dictionary of args
    :param ticker: optional - string ticker
    :param scrub_mode: optional - string type of
        scrubbing handler to run
    :param verbose: optional - bool for debugging
    """
    label = 'fetch_calls'
    datafeed_type = td_consts.DATAFEED_TD_CALLS
    exp_date = None
    latest_pricing = {}
    latest_close = None

    if work_dict:
        ticker = work_dict.get(
            'ticker',
            ticker)
        label = work_dict.get(
            'label',
            label)
        exp_date = work_dict.get(
            'exp_date',
            exp_date)
        latest_pricing = work_dict.get(
            'latest_pricing',
            latest_pricing)
        latest_close = latest_pricing.get(
            'close',
            latest_close)

    log.debug(
        f'{label} - calls - close={latest_close} '
        f'ticker={ticker}')

    exp_date = opt_dates.option_expiration().strftime(
        ae_consts.COMMON_DATE_FORMAT)
    use_url = td_consts.TD_URLS['options'].format(
        ticker,
        exp_date)
    headers = td_consts.get_auth_headers()
    session = requests.Session()
    session.headers = headers
    res = url_helper.url_helper(sess=session).get(
        use_url
    )

    if res.status_code != requests.codes.OK:
        if res.status_code in [401, 403]:
            log.critical(
                'Please check the TD_TOKEN is correct '
                f'received {res.status_code} during '
                'fetch for: calls')
        else:
            log.info(
                f'failed to get call with response={res} '
                f'code={res.status_code} '
                f'text={res.text}')
        return ae_consts.EMPTY, pd.DataFrame([{}])
    records = json.loads(res.text)
    org_records = records.get(
        'options', {}).get(
            'option', [])

    if len(org_records) == 0:
        log.info(
            'failed to get call records '
            'text={}'.format(
                res.text))
        return ae_consts.EMPTY, pd.DataFrame([{}])

    options_list = []

    # assumes UTC conversion will work with the system clock
    created_minute = (
        datetime.datetime.utcnow() - datetime.timedelta(hours=5)).strftime(
            '%Y-%m-%d %H:%M:00')
    last_close_date = ae_utils.get_last_close_str(
        fmt='%Y-%m-%d %H:%M:00')

    # hit bug where dates were None
    if not last_close_date:
        last_close_date = created_minute

    for node in org_records:
        node['date'] = last_close_date
        node['created'] = created_minute
        node['ticker'] = ticker
        if (
                node['option_type'] == 'call' and
                node['expiration_type'] == 'standard' and
                float(node['bid']) > 0.01):
            node['opt_type'] = int(ae_consts.OPTION_CALL)
            node['exp_date'] = node['expiration_date']

            new_node = {}
            for col in td_consts.TD_OPTION_COLUMNS:
                if col in node:
                    if col in td_consts.TD_EPOCH_COLUMNS:
                        # trade_date can be None
                        if node[col] == 0:
                            new_node[col] = None
                        else:
                            new_node[col] = ae_utils.epoch_to_dt(
                                epoch=node[col]/1000,
                                use_utc=False,
                                convert_to_est=True).strftime(
                                    ae_consts.COMMON_TICK_DATE_FORMAT)
                            """
                            Debug epoch ms converter:
                            """
                            """
                            print('-----------')
                            print(col)
                            print(node[col])
                            print(new_node[col])
                            print('===========')
                            """
                        # if/else valid date
                    else:
                        new_node[col] = node[col]
                    # if date column to convert
                # if column is in the row
            # convert all columns

            options_list.append(new_node)
    # end of records

    full_df = pd.DataFrame(options_list).sort_values(
        by=[
            'strike'
        ],
        ascending=True)

    num_chains = len(full_df.index)

    df = None
    if latest_close:
        df_filter = (
            (full_df['strike'] >=
                (latest_close - ae_consts.OPTIONS_LOWER_STRIKE)) &
            (full_df['strike'] <=
                (latest_close + ae_consts.OPTIONS_UPPER_STRIKE)))
        df = full_df[df_filter].copy().sort_values(
            by=[
                'date',
                'strike'
            ]).reset_index()
    else:
        mid_chain_idx = int(num_chains / 2)
        low_idx = int(
            mid_chain_idx - ae_consts.MAX_OPTIONS_LOWER_STRIKE)
        high_idx = int(
            mid_chain_idx + ae_consts.MAX_OPTIONS_UPPER_STRIKE)
        if low_idx < 0:
            low_idx = 0
        if high_idx > num_chains:
            high_idx = num_chains
        df = full_df[low_idx:high_idx].copy().sort_values(
            by=[
                'date',
                'strike'
            ]).reset_index()

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=exp_date,
        df=df)

    return ae_consts.SUCCESS, scrubbed_df
# end of fetch_calls


def fetch_puts(
        work_dict=None,
        ticker=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """fetch_puts

    Fetch Tradier option puts for a ticker and
    return a tuple: (status, ``pandas.DataFrame``)

    .. code-block:: python

        import analysis_engine.td.fetch_api as td_fetch

        puts_status, puts_df = td_fetch.fetch_puts(
            ticker='SPY')

        print(f'Fetched SPY Option Puts from Tradier status={puts_status}:')
        print(puts_df)

    :param work_dict: optional - dictionary of args
    :param ticker: optional - string ticker
    :param scrub_mode: optional - string type of
        scrubbing handler to run
    :param verbose: optional - bool for debugging
    """
    label = 'fetch_calls'
    datafeed_type = td_consts.DATAFEED_TD_PUTS
    exp_date = None
    latest_pricing = {}
    latest_close = None

    if work_dict:
        ticker = work_dict.get(
            'ticker',
            ticker)
        label = work_dict.get(
            'label',
            label)
        exp_date = work_dict.get(
            'exp_date',
            exp_date)
        latest_pricing = work_dict.get(
            'latest_pricing',
            latest_pricing)
        latest_close = latest_pricing.get(
            'close',
            latest_close)

    if verbose:
        log.info(
            f'{label} - puts - close={latest_close} '
            f'ticker={ticker}')

    exp_date = opt_dates.option_expiration().strftime(
        ae_consts.COMMON_DATE_FORMAT)
    use_url = td_consts.TD_URLS['options'].format(
        ticker,
        exp_date)
    headers = td_consts.get_auth_headers()
    session = requests.Session()
    session.headers = headers
    res = url_helper.url_helper(sess=session).get(
        use_url
    )

    if res.status_code != requests.codes.OK:
        if res.status_code in [401, 403]:
            log.critical(
                'Please check the TD_TOKEN is correct '
                f'received {res.status_code} during '
                'fetch for: puts')
        else:
            log.info(
                f'failed to get put with response={res} '
                f'code={res.status_code} '
                f'text={res.text}')
        return ae_consts.EMPTY, pd.DataFrame([{}])
    records = json.loads(res.text)
    org_records = records.get(
        'options', {}).get(
            'option', [])

    if len(org_records) == 0:
        log.info(
            'failed to get put records '
            'text={}'.format(
                res.text))
        return ae_consts.EMPTY, pd.DataFrame([{}])

    options_list = []

    # assumes UTC conversion will work with the system clock
    created_minute = (
        datetime.datetime.utcnow() - datetime.timedelta(hours=5)).strftime(
            '%Y-%m-%d %H:%M:00')
    last_close_date = ae_utils.get_last_close_str(
        fmt='%Y-%m-%d %H:%M:00')

    # hit bug where dates were None
    if not last_close_date:
        last_close_date = created_minute

    for node in org_records:
        node['date'] = last_close_date
        node['created'] = created_minute
        node['ticker'] = ticker
        if (
                node['option_type'] == 'put' and
                node['expiration_type'] == 'standard' and
                float(node['bid']) > 0.01):
            node['opt_type'] = int(ae_consts.OPTION_PUT)
            node['exp_date'] = node['expiration_date']

            new_node = {}
            for col in td_consts.TD_OPTION_COLUMNS:
                if col in node:
                    if col in td_consts.TD_EPOCH_COLUMNS:
                        # trade_date can be None
                        if node[col] == 0:
                            new_node[col] = None
                        else:
                            new_node[col] = ae_utils.epoch_to_dt(
                                epoch=node[col]/1000,
                                use_utc=False,
                                convert_to_est=True).strftime(
                                    ae_consts.COMMON_TICK_DATE_FORMAT)
                            """
                            Debug epoch ms converter:
                            """
                            """
                            print('-----------')
                            print(col)
                            print(node[col])
                            print(new_node[col])
                            print('===========')
                            """
                        # if/else valid date
                    else:
                        new_node[col] = node[col]
                    # if date column to convert
                # if column is in the row
            # convert all columns

            options_list.append(new_node)
    # end of records

    full_df = pd.DataFrame(options_list).sort_values(
        by=[
            'strike'
        ],
        ascending=True)

    num_chains = len(full_df.index)

    df = None
    if latest_close:
        df_filter = (
            (full_df['strike'] >=
                (latest_close - ae_consts.OPTIONS_LOWER_STRIKE)) &
            (full_df['strike'] <=
                (latest_close + ae_consts.OPTIONS_UPPER_STRIKE)))
        df = full_df[df_filter].copy().sort_values(
            by=[
                'date',
                'strike'
            ]).reset_index()
    else:
        mid_chain_idx = int(num_chains / 2)
        low_idx = int(
            mid_chain_idx - ae_consts.MAX_OPTIONS_LOWER_STRIKE)
        high_idx = int(
            mid_chain_idx + ae_consts.MAX_OPTIONS_UPPER_STRIKE)
        if low_idx < 0:
            low_idx = 0
        if high_idx > num_chains:
            high_idx = num_chains
        df = full_df[low_idx:high_idx].copy().sort_values(
            by=[
                'date',
                'strike'
            ]).reset_index()

    scrubbed_df = scrub_utils.ingress_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=datafeed_type,
        msg_format='df={} date_str={}',
        ds_id=ticker,
        date_str=exp_date,
        df=df)

    return ae_consts.SUCCESS, scrubbed_df
# end of fetch_puts



"""
Helper for loading derived Indicators from a local module file
"""

import os
import inspect
import types
import importlib.machinery
import uuid
import analysis_engine.consts as ae_consts
import analysis_engine.indicators.base_indicator as base_indicator
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def load_indicator_from_module(
        module_name,
        ind_dict,
        path_to_module=None,
        log_label=None,
        base_class_module_name='BaseIndicator',
        verbose=False):
    """load_indicator_from_module

    Load a custom indicator from a file

    :param module_name: string name of the indicator module
        use in to load the module
    :param path_to_module: optional - path to custom indicator file
        (default is to use the
        ``analysis_engine.indicators.base_indicator.BaseIndicator`` or
        if set the ``ind_dict['module_path']`` value)
    :param ind_dict: dictionary of keyword arguments
        to pass to the newly created derived Indicator's
        constructor
    :param log_label: optional - log tracking label
        for helping to find this indicator's logs
        (if not set the default name is the
        ``module_name`` string value)
    :param base_class_module_name: optional - string name
        for using a non-standard indicator base class
    :param verbose: optional - bool for more logging
        (default is ``False``)
    """

    default_base_module_path = ae_consts.INDICATOR_BASE_MODULE_PATH

    # modules need custom names to prevent
    # runtime collisions/module stomping
    # this allows building multiple indicator objects
    # that use the same filename but in different
    # locations on disk
    use_module_name = (
        f'{module_name}_{str(uuid.uuid4())[0:8].replace("-", "")}')

    use_log_label = log_label
    if not use_log_label:
        use_log_label = use_module_name

    if not path_to_module:
        path_to_module = ind_dict.get(
            'module_path',
            None)

    if 'verbose' not in ind_dict:
        ind_dict['verbose'] = verbose

    if not path_to_module:
        return base_indicator.BaseIndicator(
            config_dict=ind_dict,
            name=use_log_label,
            path_to_module=default_base_module_path)

    if not os.path.exists(path_to_module):
        raise Exception(
            f'{use_module_name} - did not find Indicator module at '
            f'path={path_to_module} please confirm the file exists on disk '
            'and if you are using a container, confirm it is '
            'accessible within the container')

    loader = importlib.machinery.SourceFileLoader(
        use_module_name,
        path_to_module)
    custom_indicator_module = types.ModuleType(
        loader.name)
    loader.exec_module(
        custom_indicator_module)

    found_base_object = False
    class_member_in_module = None
    for member in inspect.getmembers(custom_indicator_module):
        if module_name in str(member):
            found_base_object = True
            class_member_in_module = member
            break
    # for all members in this custom module file

    if not found_base_object:
        raise Exception(
            f'{use_module_name} load_indicator_from_module error - '
            f'did not find Indicator with base class={base_class_module_name} '
            f'from module at path={path_to_module} - please confirm '
            'the file has just one class that '
            'inherits from the base Indicator class: '
            'analysis_engine.indicators.base_indicator.BaseIndicator '
            'and try again')

        err = (
            f'{use_module_name} load_indicator_from_module error - '
            'unable to find custom indicator derived from '
            f'module={base_class_module_name} at file path={path_to_module}')
        if path_to_module:
            err = (
                f'{use_module_name} load_indicator_from_module error - '
                'analysis_engine.indicators.base_indicator.BaseIndicator '
                'was unable to find custom Indicator '
                f'module={custom_indicator_module} with provided path to \n '
                f'file: {path_to_module} \n'
                '\n'
                'Please confirm '
                'that the class inherits from the BaseIndicator '
                'class like:\n'
                '\n'
                'import analysis_engine.indicators.base_indicator '
                'as base_ind\n'
                'class MyIndicator(base_ind.BaseIndicator):\n'
                '\n'
                'If it is then please file an issue on github:\n '
                'https://github.com/AlgoTraders/stock-analysis-engine/'
                'issues/new \n\nFor now this error results in a shutdown'
                '\n')
        log.error(err)
        raise Exception(err)
    # end of if did not find the module with the correct Base Class

    if ind_dict.get('verbose', False):
        log.info(
            f'load - custom indicator module={use_module_name} '
            f'from file={path_to_module} member={class_member_in_module}')
    ind = class_member_in_module[1](
        config_dict=ind_dict,
        name=use_log_label,
        path_to_module=path_to_module)
    if ind_dict.get('verbose', False):
        log.info(
            f'ready - custom indicator={ind.__class__.__name__} from '
            f'module={use_module_name} from file={path_to_module} '
            f'member={class_member_in_module}')

    return ind
# end of load_indicator_from_module



"""
Helper for getting data from redis

Debug redis calls with:

::

    export DEBUG_REDIS=1

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json

"""

import json
import zlib
import redis
import analysis_engine.consts as ae_consts
import analysis_engine.build_result as build_result
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def get_data_from_redis_key(
        label=None,
        client=None,
        host=None,
        port=None,
        password=None,
        db=None,
        key=None,
        expire=None,
        decompress_df=False,
        serializer='json',
        encoding='utf-8'):
    """get_data_from_redis_key

    :param label: log tracking label
    :param client: initialized redis client
    :param host: not used yet - redis host
    :param port: not used yet - redis port
    :param password: not used yet - redis password
    :param db: not used yet - redis db
    :param key: not used yet - redis key
    :param expire: not used yet - redis expire
    :param decompress_df: used for decompressing
        ``pandas.DataFrame`` automatically
    :param serializer: not used yet - support for future
                       pickle objects in redis
    :param encoding: format of the encoded key in redis
    """

    decoded_data = None
    data = None

    rec = {
        'data': data
    }
    res = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    log_id = label if label else 'get-data'

    try:

        use_client = client
        if not use_client:
            log.debug(
                f'{log_id} - get key={key} new '
                f'client={host}:{port}@{db}')
            use_client = redis.Redis(
                host=host,
                port=port,
                password=password,
                db=db)
        else:
            log.debug(f'{log_id} - get key={key} client')
        # create Redis client if not set

        # https://redis-py.readthedocs.io/en/latest/index.html#redis.StrictRedis.get  # noqa
        raw_data = use_client.get(
            name=key)

        if raw_data:

            if decompress_df:
                try:
                    data = zlib.decompress(
                        raw_data).decode(
                            encoding)
                    rec['data'] = json.loads(data)

                    return build_result.build_result(
                        status=ae_consts.SUCCESS,
                        err=None,
                        rec=rec)
                except Exception as f:
                    if (
                            'while decompressing data: '
                            'incorrect header check') in str(f):
                        data = None
                        log.critical(
                            f'unable to decompress_df in redis_key={key} '
                            f'ex={f}')
                    else:
                        log.error(
                            f'failed decompress_df in redis_key={key} '
                            f'ex={f}')
                        raise f
            # allow decompression failure to fallback to previous method

            if not data:
                log.debug(f'{log_id} - decoding key={key} encoding={encoding}')
                decoded_data = raw_data.decode(encoding)

                log.debug(
                    f'{log_id} - deserial key={key} serializer={serializer}')

                if serializer == 'json':
                    data = json.loads(decoded_data)
                elif serializer == 'df':
                    data = decoded_data
                else:
                    data = decoded_data

                if data:
                    if ae_consts.ev('DEBUG_REDIS', '0') == '1':
                        log.info(
                            f'{log_id} - found key={key} '
                            f'data={ae_consts.ppj(data)}')
                    else:
                        log.debug(f'{log_id} - found key={key}')
            # log snippet - if data

            rec['data'] = data

            return build_result.build_result(
                status=ae_consts.SUCCESS,
                err=None,
                rec=rec)
        else:
            log.debug(f'{log_id} - no data key={key}')
            return build_result.build_result(
                status=ae_consts.SUCCESS,
                err=None,
                rec=rec)
    except Exception as e:
        err = (
            f'{log_id} failed - redis get from decoded={decoded_data} '
            f'data={data} key={key} ex={e}')
        log.error(err)
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=rec)
    # end of try/ex for getting redis data

    return res
# end of get_data_from_redis_key



"""
Extract an TD dataset from Redis (S3 support coming soon) and
load it into a ``pandas.DataFrame``

Supported environment variables:

::

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json

"""

import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.dataset_scrub_utils as scrub_utils
import analysis_engine.get_data_from_redis_key as redis_get
import analysis_engine.td.consts as td_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def extract_option_calls_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_option_calls_dataset

    Extract the TD options calls for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.td.extract_df_from_redis as td_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        calls_status, calls_df = td_extract.extract_option_calls_dataset(
            ticker='SPY')
        print(calls_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: optional - string type of
        scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    label = 'extract_td_calls'
    latest_close_date = ae_utils.get_last_close_str()
    use_date = date
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = f'{work_dict.get("label", label)}'
    if not use_date:
        use_date = latest_close_date

    ds_id = ticker
    df_type = td_consts.DATAFEED_TD_CALLS
    df_str = td_consts.get_datafeed_str_td(df_type=df_type)
    redis_db = ae_consts.REDIS_DB
    redis_key = f'{ticker}_{use_date}_tdcalls'
    redis_host, redis_port = ae_consts.get_redis_host_and_port(
        req=work_dict)
    redis_password = ae_consts.REDIS_PASSWORD
    s3_key = redis_key

    if work_dict:
        redis_key = work_dict.get(
            'redis_key',
            redis_key)
        redis_db = work_dict.get(
            'redis_db',
            redis_db)
        redis_password = work_dict.get(
            'redis_password',
            redis_password)
        s3_key = work_dict.get(
            's3_key',
            s3_key)
        verbose = work_dict.get(
            'verbose_td',
            verbose)

    if verbose:
        log.info(
            f'{label} - {df_str} - start - redis_key={redis_key} '
            f's3_key={s3_key}')

    exp_date_str = None
    calls_df = None
    status = ae_consts.NOT_RUN
    try:
        redis_rec = redis_get.get_data_from_redis_key(
            label=label,
            host=redis_host,
            port=redis_port,
            db=redis_db,
            password=redis_password,
            key=redis_key,
            decompress_df=True)

        status = redis_rec['status']
        if verbose:
            log.info(
                f'{label} - {df_str} redis get data key={redis_key} '
                f'status={ae_consts.get_status(status=status)}')

        if status == ae_consts.SUCCESS:
            calls_json = None
            if 'calls' in redis_rec['rec']['data']:
                calls_json = redis_rec['rec']['data']['calls']
            else:
                calls_json = redis_rec['rec']['data']
            if verbose:
                log.info(f'{label} - {df_str} redis convert calls to df')
            exp_date_str = None
            try:
                calls_df = pd.read_json(
                    calls_json,
                    orient='records')
                if len(calls_df.index) == 0:
                    return ae_consts.SUCCESS, pd.DataFrame([])
                if 'date' not in calls_df:
                    log.debug(
                        'failed to find date column in TD calls '
                        f'df={calls_df} from lens={len(calls_df.index)}')
                    return ae_consts.EMPTY, pd.DataFrame([])
                calls_df.sort_values(
                        by=[
                            'date',
                            'strike'
                        ])
                """
                for i, r in calls_df.iterrows():
                    print(r['date'])
                convert_epochs = [
                    'ask_date',
                    'bid_date',
                    'trade_date'
                ]
                for c in convert_epochs:
                    if c in calls_df:
                        calls_df[c] = pd.DatetimeIndex(pd.to_datetime(
                            calls_df[c],
                            format=ae_consts.COMMON_TICK_DATE_FORMAT
                        )).tz_localize(
                            'UTC').tz_convert(
                                'US/Eastern')
                # dates converted
                """
                exp_date_str = (
                    calls_df['exp_date'].iloc[-1])

                calls_df['date'] = calls_df['date'].dt.strftime(
                    ae_consts.COMMON_TICK_DATE_FORMAT)

            except Exception as f:
                not_fixed = True
                if (
                        'Can only use .dt accessor with '
                        'datetimelike values') in str(f):
                    try:
                        log.critical(
                            f'fixing dates in {redis_key}')
                        # remove epoch second data and
                        # use only the millisecond date values
                        bad_date = ae_consts.EPOCH_MINIMUM_DATE
                        calls_df['date'][
                            calls_df['date'] < bad_date] = None
                        calls_df = calls_df.dropna(axis=0, how='any')
                        fmt = ae_consts.COMMON_TICK_DATE_FORMAT
                        calls_df['date'] = pd.to_datetime(
                            calls_df['date'],
                            unit='ms').dt.strftime(fmt)
                        not_fixed = False
                    except Exception as g:
                        log.critical(
                            f'failed to parse date column {calls_df["date"]} '
                            f'with dt.strftime ex={f} and EPOCH EX={g}')
                        return ae_consts.EMPTY, pd.DataFrame([])
                # if able to fix error or not

                if not_fixed:
                    log.error(
                        f'{label} - {df_str} redis_key={redis_key} '
                        f'no calls df found or ex={f}')
                    return ae_consts.EMPTY, pd.DataFrame([])
                # if unable to fix - return out

                log.error(
                    f'{label} - {df_str} redis_key={redis_key} '
                    f'no calls df found or ex={f}')
                return ae_consts.EMPTY, pd.DataFrame([])
            # end of try/ex to convert to df
            if verbose:
                log.info(
                    f'{label} - {df_str} redis_key={redis_key} '
                    f'calls={len(calls_df.index)} exp_date={exp_date_str}')
        else:
            if verbose:
                log.info(
                    f'{label} - {df_str} did not find valid redis '
                    f'option calls in redis_key={redis_key} '
                    f'status={ae_consts.get_status(status=status)}')

    except Exception as e:
        if verbose:
            log.error(
                f'{label} - {df_str} - ds_id={ds_id} failed getting option '
                f'calls from redis={redis_host}:{redis_port}@{redis_db} '
                f'key={redis_key} ex={e}')
        return ae_consts.ERR, pd.DataFrame([])
    # end of try/ex extract from redis

    if verbose:
        log.info(
            f'{label} - {df_str} ds_id={ds_id} extract scrub={scrub_mode}')

    scrubbed_df = scrub_utils.extract_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=df_type,
        msg_format='df={} date_str={}',
        ds_id=ds_id,
        df=calls_df)

    status = ae_consts.SUCCESS

    return status, scrubbed_df
# end of extract_option_calls_dataset


def extract_option_puts_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_option_puts_dataset

    Extract the TD options puts for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.td.extract_df_from_redis as td_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        puts_status, puts_df = td_extract.extract_option_puts_dataset(
            ticker='SPY')
        print(puts_df)

    :param work_dict: dictionary of args
    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: optional - string type of
        scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    label = 'extract_td_puts'
    latest_close_date = ae_utils.get_last_close_str()
    use_date = date
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
        label = f'{work_dict.get("label", label)}'
    if not use_date:
        use_date = latest_close_date

    ds_id = ticker
    df_type = td_consts.DATAFEED_TD_PUTS
    df_str = td_consts.get_datafeed_str_td(df_type=df_type)
    redis_db = ae_consts.REDIS_DB
    redis_key = f'{ticker}_{use_date}_tdputs'
    redis_host, redis_port = ae_consts.get_redis_host_and_port(
        req=work_dict)
    redis_password = ae_consts.REDIS_PASSWORD
    s3_key = redis_key

    if work_dict:
        redis_db = work_dict.get(
            'redis_db',
            redis_db)
        redis_password = work_dict.get(
            'redis_password',
            redis_password)
        s3_key = work_dict.get(
            's3_key',
            s3_key)
        verbose = work_dict.get(
            'verbose_td',
            verbose)

    if verbose:
        log.info(
            f'{label} - {df_str} - start - redis_key={redis_key} '
            f's3_key={s3_key}')

    exp_date_str = None
    puts_df = None
    status = ae_consts.NOT_RUN
    try:
        redis_rec = redis_get.get_data_from_redis_key(
            label=label,
            host=redis_host,
            port=redis_port,
            db=redis_db,
            password=redis_password,
            key=redis_key,
            decompress_df=True)

        status = redis_rec['status']
        if verbose:
            log.info(
                f'{label} - {df_str} redis get data key={redis_key} '
                f'status={ae_consts.get_status(status=status)}')

        if status == ae_consts.SUCCESS:
            puts_json = None
            if 'puts' in redis_rec['rec']['data']:
                puts_json = redis_rec['rec']['data']['puts']
            else:
                puts_json = redis_rec['rec']['data']
            if verbose:
                log.info(f'{label} - {df_str} redis convert puts to df')
            try:
                puts_df = pd.read_json(
                    puts_json,
                    orient='records')
                if len(puts_df.index) == 0:
                    return ae_consts.SUCCESS, pd.DataFrame([])
                if 'date' not in puts_df:
                    log.debug(
                        'failed to find date column in TD puts '
                        f'df={puts_df} len={len(puts_df.index)}')
                    return ae_consts.EMPTY, pd.DataFrame([])
                puts_df.sort_values(
                        by=[
                            'date',
                            'strike'
                        ])
                """
                for i, r in calls_df.iterrows():
                    print(r['date'])
                convert_epochs = [
                    'ask_date',
                    'bid_date',
                    'trade_date'
                ]
                for c in convert_epochs:
                    if c in puts_df:
                        puts_df[c] = pd.DatetimeIndex(pd.to_datetime(
                            puts_df[c],
                            format=ae_consts.COMMON_TICK_DATE_FORMAT
                        )).tz_localize(
                            'UTC').tz_convert(
                                'US/Eastern')
                # dates converted
                """
                exp_date_str = (
                    puts_df['exp_date'].iloc[-1])

                puts_df['date'] = puts_df['date'].dt.strftime(
                    ae_consts.COMMON_TICK_DATE_FORMAT)

            except Exception:
                log.debug(
                    f'{label} - {df_str} redis_key={redis_key} '
                    'no puts df found')
                return ae_consts.EMPTY, pd.DataFrame([])
            # end of try/ex to convert to df
            if verbose:
                log.info(
                    f'{label} - {df_str} redis_key={redis_key} '
                    f'puts={len(puts_df.index)} exp_date={exp_date_str}')
        else:
            if verbose:
                log.info(
                    f'{label} - {df_str} did not find valid redis '
                    f'option puts in redis_key={redis_key} '
                    f'status={ae_consts.get_status(status=status)}')

    except Exception as e:
        if verbose:
            log.error(
                f'{label} - {df_str} - ds_id={ds_id} failed getting option '
                f'puts from redis={redis_host}:{redis_port}@{redis_db} '
                f'key={redis_key} ex={e}')
        return ae_consts.ERR, pd.DataFrame([])
    # end of try/ex extract from redis

    if verbose:
        log.info(
            f'{label} - {df_str} ds_id={ds_id} extract scrub={scrub_mode}')

    scrubbed_df = scrub_utils.extract_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=df_type,
        msg_format='df={} date_str={}',
        ds_id=ds_id,
        df=puts_df)

    status = ae_consts.SUCCESS

    return status, scrubbed_df
# end of extract_option_puts_dataset



"""
TA-Lib wrappers
"""

# for unittests, allow passing the mocks into the runtime if not found
try:
    import talib as ta
except Exception:
    import analysis_engine.mocks.mock_talib as ta
# end of loading talib or mocks
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


"""
Overlap

https://mrjbq7.github.io/ta-lib/func_groups/overlap_studies.html
"""


def BBANDS(
        close,
        timeperiod=5,
        nbdevup=2,
        nbdevdn=2,
        matype=0,
        verbose=False):
    """BBANDS

    Wrapper for ta.BBANDS for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        (upperband,
         middleband,
         lowerband) = BBANDS(
            close,
            timeperiod=5,
            nbdevup=2,
            nbdevdn=2,
            matype=0)

    :return: upperband, middleband, lowerband
    :param close: close prices
    :param timeperiod: number of values
        (default is ``5``)
    :param nbdevup: float - standard deviation
        to set the upper band
        (default is ``2``)
    :param nbdevdn: float - standard deviation
        to set the lower band
        (default is ``2``)
    :param matype: moving average type
        (default is ``0`` simple moving average)
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'bbands - start')
    return ta.BBANDS(
        close,
        timeperiod=timeperiod,
        nbdevup=nbdevup,
        nbdevdn=nbdevdn,
        matype=matype)
# end of BBANDS


def EMA(
        close,
        timeperiod=30,
        verbose=False):
    """EMA

    Wrapper for ta.EMA for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = EMA(
            close,
            timeperiod=30)

    :return: float
    :param close: close prices
    :param timeperiod: number of values
        (default is ``5``)
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'ema - start')
    return ta.EMA(
        close,
        timeperiod=timeperiod)
# end of EMA


def WMA(
        close,
        timeperiod=30,
        verbose=False):
    """WMA

    Wrapper for ta.WMA for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = WMA(
            close,
            timeperiod=30)

    :return: float
    :param close: close prices
    :param timeperiod: number of values
        (default is ``5``)
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'ema - start')
    return ta.WMA(
        close,
        timeperiod=timeperiod)
# end of WMA


"""
Momentum

https://mrjbq7.github.io/ta-lib/func_groups/momentum_indicators.html
"""


def ADX(
        high=None,
        low=None,
        close=None,
        timeperiod=14,
        verbose=False):
    """ADX

    Wrapper for ta.ADX for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = ADX(
            high,
            low,
            close,
            timeperiod=14)

    :param high: high list
    :param low: low list
    :param close: close list
    :param timeperiod: number of values
        in ``high``, ``low`` and ``close``
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'adx - start')
    return ta.ADX(
        high,
        low,
        close,
        timeperiod)
# end of ADX


def MACD(
        close=None,
        fast_period=12,
        slow_period=26,
        signal_period=9,
        verbose=False):
    """MACD

    Wrapper for ta.MACD for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        (macd,
         macdsignal,
         macdhist) = MACD(
            close,
            fastperiod=12,
            slowperiod=26,
            signalperiod=9)

    :param value: list of values
        (default ``closes``)
    :param fast_period: integer fast
        line
    :param slow_period: integer slow
        line
    :param signal_period: integer signal
        line
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'macd - start')
    return ta.MACD(
        close,
        fastperiod=fast_period,
        slowperiod=slow_period,
        signalperiod=signal_period)
# end of MACD


def MFI(
        high=None,
        low=None,
        close=None,
        volume=None,
        timeperiod=None,
        verbose=False):
    """MFI

    Wrapper for ta.MFI for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = MFI(
            high,
            low,
            close,
            volume,
            timeperiod=14)

    :param high: high list
    :param low: low list
    :param close: close list
    :param timeperiod: number of values
        in ``high``, ``low`` and ``close``
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'mfi - start')
    return ta.MFI(
        high,
        low,
        close,
        volume,
        timeperiod)
# end of MFI


def MOM(
        close=None,
        timeperiod=None,
        verbose=False):
    """MOM

    Wrapper for ta.MOM for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = MOM(
            close,
            timeperiod=10)

    :param high: high list
    :param low: low list
    :param close: close list
    :param timeperiod: number of values
        in ``high``, ``low`` and ``close``
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'mom - start')
    return ta.MOM(
        close,
        timeperiod)
# end of MOM


def ROC(
        close=None,
        timeperiod=None,
        verbose=False):
    """ROC

    Wrapper for ta.ROC for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = ROC(
            close,
            timeperiod=10)

    :param close: close list
    :param timeperiod: number of values
        in ``high``, ``low`` and ``close``
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'roc - start')
    return ta.ROC(
        close,
        timeperiod)
# end of ROC


def RSI(
        close=None,
        timeperiod=None,
        verbose=False):
    """RSI

    Wrapper for ta.RSI for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = RSI(
            close,
            timeperiod=14)

    :param close: close list
    :param timeperiod: number of values
        in ``high``, ``low`` and ``close``
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'rsi - start')
    return ta.RSI(
        close,
        timeperiod)
# end of RSI


def STOCH(
        high=None,
        low=None,
        close=None,
        fastk_period=None,
        slowk_period=None,
        slowk_matype=None,
        slowd_period=None,
        slowd_matype=0,
        verbose=False):
    """STOCH

    Wrapper for ta.STOCH for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        slowk, slowd = STOCH(
            high,
            low,
            close,
            fastk_period=5,
            slowk_period=3,
            slowk_matype=0,
            slowd_period=3,
            slowd_matype=0)

    :param high: list of high values
    :param low: list of low values
    :param close: list of close values
    :param fastk_period: integer num
        of fast k sticks
    :param slowk_period: integer num
        of slow k sticks
    :param slowk_matype: integer moving
        average
        (default is ``0``)
    :param slowd_period: integer num
        of slow d sticks
    :param slowd_matype: integer moving
        average
        (default is ``0``)
    :param timeperiod: number of values
        in ``high``, ``low`` and ``close``
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'stoch - start')
    return ta.STOCH(
        high=high,
        low=low,
        close=close,
        fastk_period=fastk_period,
        slowk_period=slowk_period,
        slowk_matype=slowk_matype,
        slowd_period=slowd_period,
        slowd_matype=slowd_matype)
# end of STOCH


def STOCHF(
        high=None,
        low=None,
        close=None,
        fastk_period=None,
        fastd_period=None,
        fastd_matype=0,
        verbose=False):
    """STOCHF

    Wrapper for ta.STOCHF for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        fastk, fastd = STOCHF(
            high,
            low,
            close,
            fastk_period=5,
            fastd_period=3,
            fastd_matype=0)

    :param high: list of high values
    :param low: list of low values
    :param close: list of close values
    :param fastk_period: integer num
        of fast k sticks
    :param fastd_period: integer num
        of fast d sticks
    :param fastd_matype: integer moving
        average
        (default is ``0``)
    :param timeperiod: number of values
        in ``high``, ``low`` and ``close``
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'stoch - start')
    return ta.STOCHF(
        high=high,
        low=low,
        close=close,
        fastk_period=fastk_period,
        fastd_period=fastd_period,
        fastd_matype=fastd_matype)
# end of STOCHF


def WILLR(
        high=None,
        low=None,
        close=None,
        timeperiod=None,
        verbose=False):
    """WILLR

    Wrapper for ta.WILLR for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = WILLR(
            high,
            low,
            close,
            timeperiod=14)

    :param high: high list
    :param low: low list
    :param close: close list
    :param timeperiod: number of values
        in ``high``, ``low`` and ``close``
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'willr - start')
    return ta.WILLR(
        high,
        low,
        close,
        timeperiod)
# end of WILLR


"""
Volume

https://mrjbq7.github.io/ta-lib/func_groups/volume_indicators.html
"""


def Chaikin(
        high=None,
        low=None,
        close=None,
        volume=None,
        verbose=False):
    """Chaikin

    Wrapper for ta.AD for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = AD(
            high,
            low,
            close,
            volume)

    :param value: list of values
        (default should be ``close``)
    :param volume: list of volume values
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'chaikin - start')
    return ta.AD(
        high,
        low,
        close,
        volume)
# end of Chaikin


def ChaikinADOSC(
        high=None,
        low=None,
        close=None,
        volume=None,
        fast_period=3,
        slow_period=10,
        verbose=False):
    """ChaikinADOSC

    Wrapper for ta.ADOSC for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = ADOSC(
            high,
            low,
            close,
            volume,
            fastperiod=3,
            slowperiod=10)

    :param value: list of values
        (default should be ``close``)
    :param volume: list of volume values
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'chaikinadosc - start')
    return ta.ADOSC(
        high,
        low,
        close,
        volume,
        fast_period,
        slow_period)
# end of ChaikinADOSC


def OBV(
        value=None,
        volume=None,
        verbose=False):
    """OBV

    Wrapper for ta.OBV for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = OBV(
            close,
            volume)

    :param value: list of values
        (default should be ``close``)
    :param volume: list of volume values
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'obv - start')
    return ta.OBV(
        value,
        volume)
# end of OBV


"""
Volume

https://mrjbq7.github.io/ta-lib/func_groups/volatility_indicators.html
"""


def ATR(
        high=None,
        low=None,
        close=None,
        timeperiod=None,
        verbose=False):
    """ATR

    Wrapper for ta.ATR for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = ATR(
            high,
            low,
            close,
            timeperiod=14)

    :param value: list of values
        (default should be ``close``)
    :param volume: list of volume values
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'atr - start')
    return ta.ATR(
        high,
        low,
        close,
        timeperiod=timeperiod)
# end of ATR


def NATR(
        high=None,
        low=None,
        close=None,
        timeperiod=None,
        verbose=False):
    """NATR

    Wrapper for ta.NATR for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = NATR(
            high,
            low,
            close,
            timeperiod=14)

    :param value: list of values
        (default should be ``close``)
    :param volume: list of volume values
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'natr - start')
    return ta.NATR(
        high,
        low,
        close,
        timeperiod=timeperiod)
# end of NATR


def TRANGE(
        high=None,
        low=None,
        close=None,
        verbose=False):
    """TRANGE

    Wrapper for ta.TRANGE for running unittests
    on ci/cd tools that do not provide talib

    .. code-block:: python

        real = TRANGE(
            high,
            low,
            close)

    :param value: list of values
        (default should be ``close``)
    :param volume: list of volume values
    :param verbose: show logs
    """
    if verbose:
        log.info(
            'trange - start')
    return ta.TRANGE(
        high,
        low,
        close)
# end of TRANGE



"""
**Prepare Pricing Dataset**

Prepare dataset for analysis. This task collapses
nested json dictionaries into a
csv file with a header row and stores the output
file in s3 and redis automatically.

- if key not in redis, load the key by the same name from s3
- prepare dataset from redis key
- the dataset will be stored as a dictionary with a pandas dataframe

**Sample work_dict request for this method**

`analysis_engine.api_requests.build_prepare_dataset_request <https://
github.com/AlgoTraders/stock-analysis-engine/blob/master/
analysis_engine/api_requests.py#L300>`__

::

    work_request = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        'prepared_s3_key': s3_prepared_key,
        'prepared_s3_bucket': s3_prepared_bucket_name,
        'prepared_redis_key': redis_prepared_key,
        'ignore_columns': ignore_columns,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

.. tip:: This task uses the `analysis_engine.work_tasks.
    custom_task.CustomTask class <https://github.com/A
    lgoTraders/stock-analysis-engine/blob/master/anal
    ysis_engine/work_tasks/custom_task.py>`__ for
    task event handling.

**Supported Environment Variables**

::

    export DEBUG_PREPARE=1
    export DEBUG_RESULTS=1

"""

import datetime
import redis
import celery.task as celery_task
import analysis_engine.consts as ae_consts
import analysis_engine.build_result as build_result
import analysis_engine.api_requests as api_requests
import analysis_engine.get_data_from_redis_key as redis_get
import analysis_engine.get_task_results as get_task_results
import analysis_engine.work_tasks.custom_task as custom_task
import analysis_engine.work_tasks.publish_from_s3_to_redis as s3_to_redis
import analysis_engine.dict_to_csv as dict_to_csv
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


@celery_task(
    bind=True,
    base=custom_task.CustomTask,
    queue='prepare_pricing_dataset')
def prepare_pricing_dataset(
        self,
        work_dict):
    """prepare_pricing_dataset

    Prepare dataset for analysis. Supports loading dataset from
    s3 if not found in redis. Outputs prepared artifact as a csv
    to s3 and redis.

    :param work_dict: dictionary for key/values
    """

    label = 'prepare'

    log.info(f'task - {label} - start work_dict={work_dict}')

    initial_data = None

    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    rec = {
        'ticker': None,
        'ticker_id': None,
        's3_enabled': True,
        'redis_enabled': True,
        's3_bucket': None,
        's3_key': None,
        'redis_key': None,
        'prepared_s3_key': None,
        'prepared_s3_bucket': None,
        'prepared_redis_key': None,
        'prepared_data': None,
        'prepared_size': None,
        'initial_data': None,
        'initial_size': None,
        'ignore_columns': None,
        'updated': None
    }
    res = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    try:
        ticker = work_dict.get(
            'ticker',
            ae_consts.TICKER)
        ticker_id = int(work_dict.get(
            'ticker_id',
            ae_consts.TICKER_ID))

        if not ticker:
            res = build_result.build_result(
                status=ae_consts.ERR,
                err='missing ticker',
                rec=rec)
            return res

        label = work_dict.get(
            'label',
            label)
        s3_key = work_dict.get(
            's3_key',
            None)
        s3_bucket_name = work_dict.get(
            's3_bucket',
            'pricing')
        s3_access_key = work_dict.get(
            's3_access_key',
            ae_consts.S3_ACCESS_KEY)
        s3_secret_key = work_dict.get(
            's3_secret_key',
            ae_consts.S3_SECRET_KEY)
        s3_region_name = work_dict.get(
            's3_region_name',
            ae_consts.S3_REGION_NAME)
        s3_address = work_dict.get(
            's3_address',
            ae_consts.S3_ADDRESS)
        s3_secure = work_dict.get(
            's3_secure',
            ae_consts.S3_SECURE) == '1'
        redis_address = work_dict.get(
            'redis_address',
            ae_consts.REDIS_ADDRESS)
        redis_key = work_dict.get(
            'redis_key',
            ae_consts.REDIS_KEY)
        redis_password = work_dict.get(
            'redis_password',
            ae_consts.REDIS_PASSWORD)
        redis_db = work_dict.get(
            'redis_db',
            None)
        if not redis_db:
            redis_db = ae_consts.REDIS_DB
        redis_expire = None
        if 'redis_expire' in work_dict:
            redis_expire = work_dict.get(
                'redis_expire',
                ae_consts.REDIS_EXPIRE)
        updated = work_dict.get(
            'updated',
            datetime.datetime.utcnow().strftime(
                '%Y_%m_%d_%H_%M_%S'))
        prepared_s3_key = work_dict.get(
            'prepared_s3_key',
            f'{ticker}_{updated}.csv')
        prepared_s3_bucket = work_dict.get(
            'prepared_s3_bucket',
            'prepared')
        prepared_redis_key = work_dict.get(
            'prepared_redis_key',
            'prepared')
        ignore_columns = work_dict.get(
            'ignore_columns',
            None)
        log.info(
            f'{label} redis enabled address={redis_address}@{redis_db} '
            f'key={redis_key} '
            f'prepare_s3={prepared_s3_bucket}:{prepared_s3_key} '
            f'prepare_redis={prepared_redis_key} '
            f'ignore_columns={ignore_columns}')
        redis_host = redis_address.split(':')[0]
        redis_port = redis_address.split(':')[1]

        enable_s3 = True
        enable_redis_publish = True

        rec['ticker'] = ticker
        rec['ticker_id'] = ticker_id
        rec['s3_bucket'] = s3_bucket_name
        rec['s3_key'] = s3_key
        rec['redis_key'] = redis_key
        rec['prepared_s3_key'] = prepared_s3_key
        rec['prepared_s3_bucket'] = prepared_s3_bucket
        rec['prepared_redis_key'] = prepared_redis_key
        rec['updated'] = updated
        rec['s3_enabled'] = enable_s3
        rec['redis_enabled'] = enable_redis_publish

        try:
            log.info(
                f'{label} connecting redis={redis_host}:{redis_port} '
                f'db={redis_db} key={redis_key} '
                f'updated={updated} expire={redis_expire}')
            rc = redis.Redis(
                host=redis_host,
                port=redis_port,
                password=redis_password,
                db=redis_db)
        except Exception as e:
            err = (
                f'{label} failed - redis connection to '
                f'address={redis_address}@{redis_port} '
                f'db={redis_db} key={redis_key} ex={e}')
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        # end of try/ex for connecting to redis

        initial_data_res = redis_get.get_data_from_redis_key(
            label=label,
            client=rc,
            key=redis_key)

        log.info(
            f'{label} get redis key={redis_key} '
            f'status={ae_consts.get_status(initial_data_res["status"])} '
            f'err={initial_data_res["err"]}')

        initial_data = initial_data_res['rec'].get(
            'data',
            None)

        if enable_s3 and not initial_data:

            log.info(
                f'{label} failed to find redis_key={redis_key} trying s3 from '
                f's3_key={s3_key} s3_bucket={s3_bucket_name} '
                f's3_address={s3_address}')

            get_from_s3_req = \
                api_requests.build_publish_from_s3_to_redis_request()

            get_from_s3_req['s3_enabled'] = enable_s3
            get_from_s3_req['s3_access_key'] = s3_access_key
            get_from_s3_req['s3_secret_key'] = s3_secret_key
            get_from_s3_req['s3_region_name'] = s3_region_name
            get_from_s3_req['s3_address'] = s3_address
            get_from_s3_req['s3_secure'] = s3_secure
            get_from_s3_req['s3_key'] = s3_key
            get_from_s3_req['s3_bucket'] = s3_bucket_name
            get_from_s3_req['redis_key'] = redis_key
            get_from_s3_req['label'] = f'{label}-run_publish_from_s3_to_redis'

            log.info(f'{label} load from s3={s3_key} to redis={redis_key}')

            try:
                # run in synchronous mode:
                get_from_s3_req['celery_disabled'] = True
                task_res = s3_to_redis.run_publish_from_s3_to_redis(
                    get_from_s3_req)
                if task_res.get(
                        'status',
                        ae_consts.ERR) == ae_consts.SUCCESS:
                    log.info(
                        f'{label} loaded s3={s3_bucket_name}:{s3_key} '
                        f'to redis={redis_key} retrying')
                    initial_data_res = redis_get.get_data_from_redis_key(
                        label=label,
                        client=rc,
                        key=redis_key)

                    log.info(
                        f'{label} get redis try=2 key={redis_key} status='
                        f'{ae_consts.get_status(initial_data_res["status"])} '
                        f'err={initial_data_res["err"]}')

                    initial_data = initial_data_res['rec'].get(
                        'data',
                        None)
                else:
                    err = (
                        f'{label} ERR failed loading from '
                        f'bucket={s3_bucket_name} s3_key={s3_key} to '
                        f'redis_key={redis_key} with res={task_res}')
                    log.error(err)
                    res = build_result.build_result(
                        status=ae_consts.ERR,
                        err=err,
                        rec=rec)
                    return res
            except Exception as e:
                err = (
                    f'{label} extract from s3 and publish to redis failed '
                    f'loading data from bucket={s3_bucket_name} in '
                    f's3_key={s3_key} with publish to redis_key={redis_key} '
                    f'with ex={e}')
                log.error(err)
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=err,
                    rec=rec)
                return res
            # end of try/ex for publishing from s3->redis
        # end of if enable_s3

        if not initial_data:
            err = (
                f'{label} did not find any data to prepare in '
                f'redis_key={redis_key} or s3_key={s3_key} in '
                f'bucket={s3_bucket_name}')
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res

        initial_data_num_chars = len(str(initial_data))
        initial_size_value = None
        initial_size_str = None
        if initial_data_num_chars < ae_consts.PREPARE_DATA_MIN_SIZE:
            err = (
                f'{label} not enough data={initial_data_num_chars} in '
                f'redis_key={redis_key} or s3_key={s3_key} in '
                f'bucket={s3_bucket_name}')
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        else:
            initial_size_value = initial_data_num_chars / 1024000
            initial_size_str = ae_consts.to_f(initial_size_value)
            if ae_consts.ev('DEBUG_PREPARE', '0') == '1':
                log.info(
                    f'{label} initial - redis_key={redis_key} '
                    f'data={str(initial_data)}')
            else:
                log.info(
                    f'{label} initial - redis_key={redis_key} data '
                    f'size={initial_size_str} MB')
        # end of trying to get initial_data

        rec['initial_data'] = initial_data
        rec['initial_size'] = initial_data_num_chars

        prepare_data = None

        try:
            if ae_consts.ev('DEBUG_PREPARE', '0') == '1':
                log.info(
                    f'{label} data={ae_consts.ppj(initial_data)} - flatten - '
                    f'{initial_size_str} MB from redis_key={redis_key}')
            else:
                log.info(
                    f'{label} flatten - {initial_size_str} MB from '
                    f'redis_key={redis_key}')
            prepare_data = dict_to_csv.flatten_dict(
                data=initial_data)
        except Exception as e:
            prepare_data = None
            err = (
                f'{label} flatten - convert to csv failed with ex={e} '
                f'redis_key={redis_key}')
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        # end of try/ex

        if not prepare_data:
            err = (
                f'{label} flatten - did not return any data from '
                f'redis_key={redis_key} or s3_key={s3_key} in '
                f'bucket={s3_bucket_name}')
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        # end of prepare_data

        prepare_data_num_chars = len(str(prepare_data))
        prepare_size_value = None

        if prepare_data_num_chars < ae_consts.PREPARE_DATA_MIN_SIZE:
            err = (
                f'{label} prepare - there is not enough '
                f'data={prepare_data_num_chars} in redis_key={redis_key}')
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        else:
            prepare_size_value = prepare_data_num_chars / 1024000
            prepare_size_str = ae_consts.to_f(prepare_size_value)
            if ae_consts.ev('DEBUG_PREPARE', '0') == '1':
                log.info(
                    f'{label} data={redis_key} - prepare - '
                    f'redis_key={ae_consts.ppj(prepare_data)}')
            else:
                log.info(
                    f'{label} prepare - redis_key={redis_key} data '
                    f'size={prepare_size_str} MB')
        # end of trying to the size of the prepared data

        rec['prepared_data'] = prepare_data
        rec['prepared_size'] = prepare_data_num_chars

        res = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

        rc = None

    except Exception as e:
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=(
                'failed - prepare_pricing_dataset '
                f'dict={work_dict} with ex={e}'),
            rec=rec)
        log.error(f'{label} - {res["err"]}')
    # end of try/ex

    log.info(
        'task - prepare_pricing_dataset done - '
        f'{label} - status={ae_consts.get_status(res["status"])}')

    return get_task_results.get_task_results(
        work_dict=work_dict,
        result=res)
# end of prepare_pricing_dataset


def run_prepare_pricing_dataset(
        work_dict):
    """run_prepare_pricing_dataset

    Celery wrapper for running without celery

    :param work_dict: task data
    """

    label = work_dict.get(
        'label',
        '')

    log.info(f'run_prepare_pricing_dataset - {label} - start')

    response = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec={})
    task_res = {}

    # allow running without celery
    if ae_consts.is_celery_disabled(
            work_dict=work_dict):
        work_dict['celery_disabled'] = True
        task_res = prepare_pricing_dataset(
            work_dict)
        if task_res:
            response = task_res.get(
                'result',
                task_res)
            if ae_consts.ev('DEBUG_RESULTS', '0') == '1':
                response_details = response
                try:
                    response_details = ae_consts.ppj(response)
                except Exception:
                    response_details = response
                log.info(f'{label} task result={response_details}')
        else:
            log.error(
                f'{label} celery was disabled but the task={response} '
                'did not return anything')
        # end of if response
    else:
        task_res = prepare_pricing_dataset.delay(
            work_dict=work_dict)
        rec = {
            'task_id': task_res
        }
        response = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)
    # if celery enabled

    if response:
        if ae_consts.ev('DEBUG_RESULTS', '0') == '1':
            log.info(
                f'run_prepare_pricing_dataset - {label} - done '
                f'status={ae_consts.get_status(response["status"])} '
                f'err={response["err"]} rec={response["rec"]}')
        else:
            log.info(
                f'run_prepare_pricing_dataset - {label} - done '
                f'status={ae_consts.get_status(response["status"])} '
                f'err={response["err"]}')
    else:
        log.info(
            f'run_prepare_pricing_dataset - {label} - done '
            'no response')
    # end of if/else response

    return response
# end of run_prepare_pricing_dataset



"""
Supported Fetch calls

- Convert a FinViz Screener URL to a list of
  tickers.

"""

import requests
import bs4
import pandas as pd
import analysis_engine.build_result as req_utils
from analysis_engine.utils import get_last_close_str
from analysis_engine.consts import NOT_RUN
from analysis_engine.consts import SUCCESS
from analysis_engine.consts import ERR
from analysis_engine.consts import EX
from analysis_engine.finviz.consts import DEFAULT_FINVIZ_COLUMNS
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def fetch_tickers_from_screener(
        url,
        columns=DEFAULT_FINVIZ_COLUMNS,
        as_json=False,
        soup_selector='td.screener-body-table-nw',
        label='fz-screen-converter'):
    """fetch_tickers_from_screener

    Convert all the tickers on a FinViz screener
    url to a ``pandas.DataFrame``. Returns a dictionary
    with a ticker list and DataFrame or a json-serialized
    DataFrame in a string (by default ``as_json=False`` will
    return a ``pandas.DataFrame`` if the
    ``returned-dictionary['status'] == SUCCESS``

    Works with urls created on:

    https://finviz.com/screener.ashx

    .. code-block:: python

        import analysis_engine.finviz.fetch_api as fv

        url = (
            'https://finviz.com/screener.ashx?'
            'v=111&'
            'f=cap_midunder,exch_nyse,fa_div_o5,idx_sp500'
            '&ft=4')
        res = fv.fetch_tickers_from_screener(url=url)
        print(res)

    :param url: FinViz screener url
    :param columns: ordered header column as a list of strings
                    and corresponds to the header row from the
                    FinViz screener table
    :param soup_selector: ``bs4.BeautifulSoup.selector`` string
                          for pulling selected html data
                          (by default ``td.screener-body-table-nw``)
    :param as_json: FinViz screener url
    :param label: log tracking label string
    """

    rec = {
        'data': None,
        'created': get_last_close_str(),
        'tickers': []
    }
    res = req_utils.build_result(
        status=NOT_RUN,
        err=None,
        rec=rec)

    try:

        log.info(f'{label} fetching url={url}')

        response = requests.get(url)

        if response.status_code != requests.codes.ok:
            err = (
                f'{label} finviz returned non-ok HTTP (200) '
                f'status_code={response.status_code} with '
                f'text={response.text} for url={url}')
            log.error(err)
            return req_utils.build_result(
                status=ERR,
                err=err,
                rec=rec)
        # end of checking for a good HTTP response status code

        soup = bs4.BeautifulSoup(
            response.text,
            features='html.parser')
        selected = soup.select(soup_selector)

        log.debug(f'{label} found={len(selected)} url={url}')

        ticker_list = []
        rows = []
        use_columns = columns
        num_columns = len(use_columns)
        new_row = {}
        col_idx = 0

        for idx, node in enumerate(selected):

            if col_idx >= num_columns:
                col_idx = 0
            column_name = use_columns[col_idx]
            test_text = str(node.text).lower().strip()
            col_idx += 1

            if column_name != 'ignore' and (
                    test_text != 'save as portfolio'
                    and test_text != 'export'):

                cur_text = str(node.text).strip()

                if column_name == 'ticker':
                    ticker_list.append(cur_text)
                    new_row[column_name] = cur_text.upper()
                else:
                    new_row[column_name] = cur_text
                # end of filtering bad sections around table

                if len(new_row) >= num_columns:
                    log.debug(f'{label} adding ticker={new_row["ticker"]}')
                    rows.append(new_row)
                    new_row = {}
                    col_idx = 0
                # end of if valid row
            # end if column is valid
        # end of walking through all matched html data on the screener

        log.debug(
            f'{label} done convert url={url} to tickers={ticker_list} '
            f'rows={len(rows)}')

        df = pd.DataFrame(
            rows)

        log.info(
            f'{label} fetch done - df={len(df.index)} from url={url} '
            f'with tickers={ticker_list} rows={len(rows)}')

        rec['tickers'] = ticker_list
        rec['data'] = df

        res = req_utils.build_result(
            status=SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        rec['tickers'] = []
        rec['data'] = None
        err = (
            f'{label} failed converting screen url={url} to list with ex={e}')
        log.error(err)
        res = req_utils.build_result(
            status=EX,
            err=err,
            rec=rec)
    # end of try/ex

    return res
# end of fetch_tickers_from_screener



"""
Helper for converting a dictionary to an algorithm-ready
dataset
"""

import json
import zlib
import pandas as pd
import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def prepare_dict_for_algo(
        data,
        compress=False,
        encoding='utf-8',
        convert_to_dict=False,
        dataset_names=None):
    """prepare_dict_for_algo

    :param data: string holding contents of an algorithm-ready
        file, s3 key or redis-key
    :param compress: optional - boolean flag for decompressing
        the contents of the ``data`` if necessary
        (default is ``False`` and algorithms
        use ``zlib`` for compression)
    :param convert_to_dict: optional - bool for s3 use ``False``
        and for files use ``True``
    :param encoding: optional - string for data encoding
    :param dataset_names: optional - list of string keys
        for each dataset node in:
        ``dataset[ticker][0]['data'][dataset_names[0]]``
    """
    log.debug('start')
    use_data = None
    parsed_data = None
    data_as_dict = None

    if compress:
        log.debug('decompressing')
        parsed_data = zlib.decompress(
            data).decode(
                encoding)
    else:
        parsed_data = data

    if not parsed_data:
        log.error('failed parsing')
        return None

    log.debug('loading as dict')
    use_data = {}
    if convert_to_dict:
        data_as_dict = json.loads(parsed_data)
    else:
        data_as_dict = parsed_data
    if len(data_as_dict) == 0:
        log.error(
            'empty algorithm-ready dictionary')
        return use_data

    empty_pd = pd.DataFrame([{}])

    use_serialized_datasets = dataset_names
    if not use_serialized_datasets:
        use_serialized_datasets = ae_consts.DEFAULT_SERIALIZED_DATASETS
    log.info(f'converting serialized_datasets={use_serialized_datasets}')
    num_datasets = 0
    for ticker in data_as_dict:
        if ticker not in use_data:
            use_data[ticker] = []
        for node in data_as_dict[ticker]:
            new_node = {
                'id': node['id'],
                'date': node['date'],
                'data': {}
            }
            for ds_key in node['data']:
                if ds_key in use_serialized_datasets:
                    new_node['data'][ds_key] = empty_pd
                    if node['data'][ds_key]:
                        new_node['data'][ds_key] = pd.read_json(
                            node['data'][ds_key],
                            orient='records')
                        num_datasets += 1
                # if supported dataset key
            # end for all datasets in this node
            use_data[ticker].append(new_node)
        # end for all datasets on this date to load
    # end for all tickers in the dataset

    if num_datasets:
        log.info(f'found datasets={num_datasets}')
    else:
        log.error(f'did not find any datasets={num_datasets}')

    return use_data
# end of prepare_dict_for_algo



#!/usr/bin/env python

"""
Train a DNN from a trading history

::

    train_from_history.py -b s3_bucket -k s3_key
"""

import argparse
import datetime
import numpy as np
import numpy.random as np_random
import pandas as pd
import pandas.api.types as pandas_types
import sklearn.model_selection as tt_split
import keras.wrappers.scikit_learn as keras_scikit
import tensorflow as tf
import analysis_engine.consts as ae_consts
import analysis_engine.load_history_dataset as load_history
import analysis_engine.ai.build_regression_dnn as build_dnn
import analysis_engine.ai.build_datasets_using_scalers as build_scaler_datasets
import analysis_engine.ai.build_scaler_dataset_from_df as build_scaler_df
import analysis_engine.ai.plot_dnn_fit_history as plot_fit_history
import analysis_engine.plot_trading_history as plot_trading_history
import spylunking.log.setup_logging as log_utils
from copy import deepcopy

# ensure reproducible results
# machinelearningmastery.com/reproducible-results-neural-networks-keras/
np_random.seed(1)

log = log_utils.build_colorized_logger(
    name='train-dnn-from-history')

choices = ['close', 'high', 'low']


def train_and_predict_from_history_in_s3():
    """train_and_predict_from_history_in_s3

    Run a derived algorithm with an algorithm config dictionary

    :param config_dict: algorithm config dictionary
    """

    log.debug('start - plot')

    parser = argparse.ArgumentParser(
        description=(
            'train a dnn to predict a column from a'
            'a trading history file in s3'))
    parser.add_argument(
        '-b',
        help=(
            's3 bucket'),
        required=False,
        dest='s3_bucket')
    parser.add_argument(
        '-k',
        help=(
            's3 key'),
        required=False,
        dest='s3_key')
    parser.add_argument(
        '-q',
        help=(
            'disable scaler normalization and '
            'only use 2 of close, high, or low with open to '
            'predict the remainder of close, high, or low'),
        required=False,
        dest='disable_scaler',
        action='store_true')
    parser.add_argument(
        '-c',
        help=(
            'column(s) to predict'),
        required=False,
        dest='predict_features',
        choices=choices,
        nargs='*')
    parser.add_argument(
        '-n',
        help=(
            'number of DNNs to create'),
        required=False,
        dest='number_of_dnns',
        default=1,
        type=int)
    parser.add_argument(
        '-s',
        help=(
            'send plots to slack'),
        required=False,
        dest='send_plots_to_slack',
        action='store_true')
    parser.add_argument(
        '-d',
        help=(
            'debug'),
        required=False,
        dest='debug',
        action='store_true')
    args = parser.parse_args()

    use_scalers = True
    s3_access_key = ae_consts.S3_ACCESS_KEY
    s3_secret_key = ae_consts.S3_SECRET_KEY
    s3_region_name = ae_consts.S3_REGION_NAME
    s3_address = ae_consts.S3_ADDRESS
    s3_secure = ae_consts.S3_SECURE
    compress = True

    s3_bucket = (
        f'algohistory')
    s3_key = (
        f'algo_training_SPY.json')
    predict_features = ['close']
    number_of_dnns = 1
    send_plots_to_slack = False

    debug = False

    if args.s3_bucket:
        s3_bucket = args.s3_bucket
    if args.s3_key:
        s3_key = args.s3_key
    if args.disable_scaler:
        use_scalers = False
    if args.debug:
        debug = True
    if args.predict_features and len(args.predict_features):
        predict_features = set(args.predict_features)  # remove any duplicates
    if args.number_of_dnns != number_of_dnns:
        number_of_dnns = args.number_of_dnns
    if args.send_plots_to_slack:
        send_plots_to_slack = args.send_plots_to_slack

    load_res = load_history.load_history_dataset(
        s3_enabled=True,
        s3_key=s3_key,
        s3_address=s3_address,
        s3_bucket=s3_bucket,
        s3_access_key=s3_access_key,
        s3_secret_key=s3_secret_key,
        s3_region_name=s3_region_name,
        s3_secure=s3_secure,
        compress=compress)

    algo_config = load_res.get(
        'algo_config_dict',
        None)
    algo_name = load_res.get(
        'algo_name',
        None)
    tickers = load_res.get(
        'tickers',
        [
            'SPY',
        ])
    ticker = tickers[0]

    log.info(
        f'found algo: {algo_name}')

    if debug:
        log.info(
            f'config: {ae_consts.ppj(algo_config)}')

    df = load_res[ticker]
    df['date'] = pd.to_datetime(
        df['date'])
    df['minute'] = pd.to_datetime(
        df['minute'])
    ticker = df['ticker'].iloc[0]

    dnn_config = {
        'layers': [
            {
                'num_neurons': 150,
                'init': 'uniform',
                'activation': 'relu'
            },
            {
                'num_neurons': 100,
                'init': 'uniform',
                'activation': 'relu'
            },
            {
                'num_neurons': 50,
                'init': 'uniform',
                'activation': 'relu'
            },
            {
                'num_neurons': 1,
                'init': 'uniform',
                'activation': 'relu'
            }
        ]
    }

    compile_config = {
        'loss': 'mse',
        'optimizer': 'adam',
        'metrics': [
            'accuracy',
            'mse',
            'mae',
            'mape',
            'cosine'
        ]
    }

    for predict_feature in predict_features:
        for index in range(number_of_dnns):
            log.info(f'Creating DNN-{index+1} for column: {predict_feature}')
            create_column_dnn(
                predict_feature=predict_feature,
                ticker=ticker,
                debug=debug,
                use_scalers=use_scalers,
                df=deepcopy(df),
                dnn_config=deepcopy(dnn_config),
                compile_config=compile_config,
                s3_bucket=s3_bucket,
                s3_key=s3_key,
                send_plots_to_slack=send_plots_to_slack)
# end of train_and_predict_from_history_in_s3


def create_column_dnn(
    predict_feature='close',
    ticker='',
    debug=False,
    use_epochs=10,
    use_batch_size=10,
    use_test_size=0.1,
    use_random_state=1,
    use_seed=7,
    use_shuffle=False,
    model_verbose=True,
    fit_verbose=True,
    use_scalers=True,
    df=[],
    dnn_config={},
    compile_config={},
    s3_bucket='',
    s3_key='',
        send_plots_to_slack=False):
    """create_column_dnn

    For scaler-normalized datasets this will
    compile numeric columns and ignore string/non-numeric
    columns as training and test feature columns

    :param predict_feature: Column to create DNN with
    :param ticker: Ticker being used
    :param debug: Debug mode
    :param use_epochs: Epochs times to use
    :param use_batch_size: Batch size to use
    :param use_test_size: Test size to use
    :param use_random_state: Random state to train with
    :param use_seed: Seed used to build scalar datasets
    :param use_shuffle: To shuffle the regression estimator or not
    :param model_verbose: To use a verbose Keras regression model or not
    :param fit_verbose: To use a verbose fitting of the regression estimator
    :param use_scalers: To build using scalars or not
    :param df: Ticker dataset
    :param dnn_config: Deep Neural Net keras model json to build the model
    :param compile_config: Deep Neural Net dictionary of compile options
    :param s3_bucket: S3 Bucket
    :param s3_key: S3 Key
    """

    df_filter = (df[f'{predict_feature}'] >= 0.1)
    first_date = df[df_filter]['date'].iloc[0]
    end_date = df[df_filter]['date'].iloc[-1]

    if 'minute' in df:
        found_valid_minute = df['minute'].iloc[0]
        if found_valid_minute:
            first_date = df[df_filter]['minute'].iloc[0]
            end_date = df[df_filter]['minute'].iloc[-1]

    num_rows = len(df.index)
    log.info(
        f'prepared training data from '
        f'history {s3_bucket}@{s3_key} '
        f'rows={num_rows} '
        f'dates: {first_date} to {end_date}')

    if debug:
        for i, r in df.iterrows():
            log.info(
                f'{r["minute"]} - {r["{}".format(predict_feature)]}')
        # end of for loop

        log.info(
            f'columns: {df.columns.values}')
        log.info(
            f'rows: {len(df.index)}')
    # end of debug

    use_all_features = use_scalers
    all_features = []
    train_features = []
    if use_all_features:
        for c in df.columns.values:
            if (
                    pandas_types.is_numeric_dtype(df[c]) and
                    c not in train_features):
                if c != predict_feature:
                    train_features.append(c)
                if c not in all_features:
                    all_features.append(c)

        dnn_config['layers'][-1]['activation'] = (
            'sigmoid')
    else:
        temp_choices = choices[:]
        temp_choices.remove(predict_feature)
        train_features = ['open']
        train_features.extend(temp_choices)
        all_features = [
            f'{predict_feature}'
        ] + train_features

    num_features = len(train_features)
    features_and_minute = [
        'minute'
    ] + all_features

    log.info(
        f'converting columns to floats')

    timeseries_df = df[df_filter][features_and_minute].fillna(-10000.0)
    converted_df = timeseries_df[all_features].astype('float32')

    train_df = None
    test_df = None
    scaler_predictions = None
    if use_all_features:
        scaler_res = build_scaler_datasets.build_datasets_using_scalers(
            train_features=train_features,
            test_feature=predict_feature,
            df=converted_df,
            test_size=use_test_size,
            seed=use_seed)
        if scaler_res['status'] != ae_consts.SUCCESS:
            log.error(
                'failed to build scaler train and test datasets')
            return
        train_df = scaler_res['scaled_train_df']
        test_df = scaler_res['scaled_test_df']
        x_train = scaler_res['x_train']
        x_test = scaler_res['x_test']
        y_train = scaler_res['y_train']
        y_test = scaler_res['y_test']
        scaler_predictions = scaler_res['scaler_test']
    else:
        log.info(
            f'building train and test dfs from subset of features')
        train_df = converted_df[train_features]
        test_df = converted_df[[predict_feature]]

        log.info(
            f'splitting {num_rows} into test and training '
            f'size={use_test_size}')

        (x_train,
         x_test,
         y_train,
         y_test) = tt_split.train_test_split(
            train_df,
            test_df,
            test_size=use_test_size,
            random_state=use_random_state)

    log.info(
        f'split breakdown - '
        f'x_train={len(x_train)} '
        f'x_test={len(x_test)} '
        f'y_train={len(y_train)} '
        f'y_test={len(y_test)}')

    def set_model():
        return build_dnn.build_regression_dnn(
            num_features=num_features,
            compile_config=compile_config,
            model_config=dnn_config)

    estimator = keras_scikit.KerasRegressor(
        build_fn=set_model,
        epochs=use_epochs,
        batch_size=use_batch_size,
        verbose=model_verbose)

    log.info(
        f'fitting estimator - '
        f'predicting={predict_feature} '
        f'epochs={use_epochs} '
        f'batch={use_batch_size} '
        f'test_size={use_test_size} '
        f'seed={use_seed}')

    history = estimator.fit(
        x_train,
        y_train,
        validation_data=(
            x_train,
            y_train),
        epochs=use_epochs,
        batch_size=use_batch_size,
        shuffle=use_shuffle,
        verbose=fit_verbose)

    created_on = (
        datetime.datetime.now().strftime(
            ae_consts.COMMON_TICK_DATE_FORMAT))
    plot_fit_history.plot_dnn_fit_history(
        df=history.history,
        title=(
            f'DNN Errors Over Training Epochs\n'
            f'Training Data: s3://{s3_bucket}/{s3_key}\n'
            f'Created: {created_on}'),
        red='mean_squared_error',
        blue='mean_absolute_error',
        green='acc',
        orange='cosine_proximity',
        send_plots_to_slack=send_plots_to_slack)

    # on production use newly fetched pricing data
    # not the training data
    predict_records = []
    if use_all_features:
        prediction_res = build_scaler_df.build_scaler_dataset_from_df(
            df=converted_df[train_features])
        if prediction_res['status'] == ae_consts.SUCCESS:
            predict_records = prediction_res['df']
    else:
        predict_records = converted_df[train_features]

    log.info(
        f'making predictions: {len(predict_records)}')

    predictions = estimator.model.predict(
        predict_records,
        verbose=True)

    np.set_printoptions(threshold=np.nan)
    indexes = tf.argmax(predictions, axis=1)
    data = {}
    data['indexes'] = indexes
    price_predictions = []
    if use_all_features and scaler_predictions:
        price_predictions = [
            ae_consts.to_f(x) for x in
            scaler_predictions.inverse_transform(
                predictions.reshape(-1, 1)).reshape(-1)]
    else:
        price_predictions = [ae_consts.to_f(x[0]) for x in predictions]

    timeseries_df[f'predicted_{predict_feature}'] = price_predictions
    timeseries_df['error'] = (
        timeseries_df[f'{predict_feature}'] -
        timeseries_df[f'predicted_{predict_feature}'])

    output_features = [
        'minute',
        f'{predict_feature}',
        f'predicted_{predict_feature}',
        'error'
    ]

    date_str = (
        f'Dates: {timeseries_df["minute"].iloc[0]} '
        f'to '
        f'{timeseries_df["minute"].iloc[-1]}')

    log.info(
        f'historical {predict_feature} with predicted {predict_feature}: '
        f'{timeseries_df[output_features]}')
    log.info(
        date_str)
    log.info(
        f'Columns: {output_features}')

    average_error = ae_consts.to_f(
        timeseries_df['error'].sum() / len(timeseries_df.index))

    log.info(
        f'Average historical {predict_feature} '
        f'vs predicted {predict_feature} error: '
        f'{average_error}')

    log.info(
        f'plotting historical {predict_feature} vs predicted {predict_feature}'
        f' from training with columns={num_features}')

    ts_filter = (timeseries_df[f'{predict_feature}'] > 0.1)
    latest_feature = (
        timeseries_df[ts_filter][f'{predict_feature}'].iloc[-1])
    latest_predicted_feature = (
        timeseries_df[ts_filter][f'predicted_{predict_feature}'].iloc[-1])

    log.info(
        f'{end_date} {predict_feature}={latest_feature} '
        f'with '
        f'predicted_{predict_feature}={latest_predicted_feature}')

    plot_trading_history.plot_trading_history(
        title=(
            f'{ticker} - Historical {predict_feature.title()} vs '
            f'Predicted {predict_feature.title()}\n'
            f'Number of Training Features: {num_features}\n'
            f'{date_str}'),
        df=timeseries_df,
        red=f'{predict_feature}',
        blue=f'predicted_{predict_feature}',
        green=None,
        orange=None,
        date_col='minute',
        date_format='%d %H:%M:%S\n%b',
        xlabel='minute',
        ylabel=(
            f'Historical {predict_feature.title()} vs '
            f'Predicted {predict_feature.title()}'),
        df_filter=ts_filter,
        width=8.0,
        height=8.0,
        show_plot=True,
        dropna_for_all=False,
        send_plots_to_slack=send_plots_to_slack)
# end of create_column_dnn


if __name__ == '__main__':
    train_and_predict_from_history_in_s3()



"""
Indicator Processor

- v1 Indicator type: ``supported``
    Binary decision support on buys and sells
    This is like an alert threshold that is ``on`` or ``off``

- v2 Indicator type: ``not supported``
    Support for buy or sell value range
    This is like an alert threshold between a ``lower``
    and ``upper`` bound
"""

import os
import json
import analysis_engine.consts as ae_consts
import analysis_engine.indicators.build_indicator_node as build_indicator
import analysis_engine.indicators.load_indicator_from_module as load_indicator
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


class IndicatorProcessor:
    """IndicatorProcessor"""

    def __init__(
            self,
            config_dict,
            config_file=None,
            ticker=None,
            label=None,
            verbose=False,
            verbose_indicators=False):
        """__init__

        Algorithm's use the ``IndicatorProcessor`` to drive
        how the underlying indicators are created and configured
        to determine buy and sell conditions. Create an
        IndicatorProcessor by passing in a valid:

        ``config_dict`` or a path to a local `config_file``

        Please refer to the `included algorithm config file <http
        s://github.com/AlgoTraders/stock-analysis-engine/blob/mas
        ter/tests/algo_configs/test_5_days_ahead.json>`__ for
        more details on how to create your own.

        :param config_dict: - dictionary for creating
            indicators and rules for buy/sell conditions
            and parameters for each indicator
        :param config_file: path to a json file
            containing custom algorithm object
            member values (like indicator configuration and
            predict future date units ahead for a backtest)
        :param ticker: optional - single ticker string
            indicators should focus on math, fundamentals,
            sentiment and other data, but the context about
            which ticker this is for should hopefully be
            abstracted from how an indicator predicts
            buy and sell conditions
        :param label: optional - string log tracking
            this class in the logs (usually just the algo
            name is good enough to help debug issues
            when running distributed)
        :param verbose: optional - bool for more logging
            (default is ``False``)
        :param verbose_indicators: optional - bool for more logging
            for all indicators managed by this ``IndicatorProcessor``
            (default is ``False``)
        """

        self.config_dict = config_dict
        if not self.config_dict:
            if config_file:
                if not os.path.exists(config_file):
                    raise Exception(
                        f'Unable to find config_file: {config_file}')
                # end of if file does not exist on the disk
                self.config_dict = json.loads(
                    open(config_file, 'r').read())
        # end of trying to ensure the config_dict is ready

        if not self.config_dict:
            raise Exception(
                'Missing either a config_dict or a config_file to '
                'create the IndicatorProcessor')

        self.last_ind_obj = None
        self.ticker = ticker
        self.ind_dict = {}
        self.num_indicators = len(self.config_dict.get(
            'indicators',
            []))
        self.label = label
        if not self.label:
            self.label = 'idprc'

        self.latest_report = {}
        self.reports = []

        self.verbose = verbose
        self.verbose_indicators = verbose_indicators

        if not self.verbose_indicators:
            self.verbose_indicators = self.config_dict.get(
                'verbose_indicators',
                False)

        self.build_indicators_for_config(
            config_dict=self.config_dict)
    # end of __init__

    def get_last_ind_obj(
            self):
        """get_last_ind_obj"""
        return self.last_ind_obj
    # end of get_last_ind_obj

    def get_latest_report(
            self,
            algo_id=None,
            ticker=None,
            dataset=None):
        """get_latest_report

        Return the latest report as a method that can be
        customized by a derived class from the
        ``IndicatorProcessor``

        :param algo_id: optional - string -
            algo identifier label for debugging datasets
            during specific dates
        :param ticker: optional - string - ticker
        :param dataset: optional - a dictionary of
            identifiers (for debugging) and
            multiple pandas ``pd.DataFrame`` objects. Dictionary where keys
            represent a label from one of the data sources (``IEX``,
            ``Yahoo``, ``FinViz`` or other). Here is the supported
            dataset structure for the process method:
        """

        return self.latest_report
    # end of get_latest_report

    def get_num_indicators(
            self):
        """get_num_indicators"""
        return self.num_indicators
    # end of get_num_indicators

    def get_label(
            self):
        """get_label"""
        return self.label
    # end of get_label

    def get_indicators(
            self):
        """get_indicators"""
        return self.ind_dict
    # end of get_indicators

    def build_indicators_for_config(
            self,
            config_dict):
        """build_indicators_for_config

        Convert the dictionary into an internal dictionary
        for quickly processing results

        :param config_dict: initailized algorithm config
            dictionary
        """

        if 'indicators' not in config_dict:
            log.error('missing "indicators" list in the config_dict')
            return

        if self.verbose:
            log.info(
                f'{self.label} start - '
                f'building indicators={self.num_indicators}')

        for idx, node in enumerate(config_dict['indicators']):
            percent_done = ae_consts.get_percent_done(
                progress=(idx + 1),
                total=self.num_indicators)
            percent_label = (
                f'ticker={self.ticker} {percent_done} '
                f'{idx+1}/{self.num_indicators}')
            # this will throw on errors parsing to make
            # it easeir to debug
            # before starting the algo and waiting for an error
            # in the middle of a backtest
            new_node = build_indicator.build_indicator_node(
                node=node)
            if new_node:
                indicator_key_name = new_node['report']['name']
                if self.verbose:
                    log.info(
                        f'{self.label} - '
                        f'preparing indicator={indicator_key_name} '
                        f'node={new_node} {percent_label}')
                else:
                    log.debug(
                        f'{self.label} - '
                        f'preparing indicator={indicator_key_name} '
                        f'{percent_label}')
                self.ind_dict[indicator_key_name] = new_node
                self.ind_dict[indicator_key_name]['obj'] = None

                base_class_indicator = node.get(
                    'base_class',
                    'BaseIndicator')

                self.ind_dict[indicator_key_name]['obj'] = \
                    load_indicator.load_indicator_from_module(
                        module_name=new_node['report']['module_name'],
                        path_to_module=new_node['report']['path_to_module'],
                        ind_dict=new_node,
                        log_label=indicator_key_name,
                        base_class_module_name=base_class_indicator,
                        verbose=self.verbose_indicators)

                log.debug(
                    f'{self.label} - '
                    f'created indicator={indicator_key_name} '
                    f'{percent_label}')
            else:
                raise Exception(
                    f'{self.label} - '
                    f'failed creating indicator {idx} node={node}')
        # end for all indicators in the config

        if self.verbose:
            log.info(
                f'{self.label} done - '
                f'built={len(self.ind_dict)} '
                f'from indicators={self.num_indicators}')
    # end of build_indicators_for_config

    def process(
            self,
            algo_id,
            ticker,
            dataset):
        """process

        :param algo_id: string - algo identifier label for debugging datasets
            during specific dates
        :param ticker: string - ticker
        :param dataset: a dictionary of identifiers (for debugging) and
            multiple pandas ``pd.DataFrame`` objects. Dictionary where keys
            represent a label from one of the data sources (``IEX``,
            ``Yahoo``, ``FinViz`` or other). Here is the supported
            dataset structure for the process method:
        """
        self.latest_report = {
            'id': algo_id,
            'ticker': ticker,
            'buys': [],
            'sells': [],
            'num_indicators': self.num_indicators,
            'date': dataset.get('date', None)
        }
        for idx, ind_id in enumerate(self.ind_dict):
            ind_node = self.ind_dict[ind_id]
            ind_obj = ind_node['obj']
            percent_done = ae_consts.get_percent_done(
                progress=(idx + 1),
                total=self.num_indicators)
            percent_label = (
                f'ticker={self.ticker} {percent_done} '
                f'{idx+1}/{self.num_indicators}')
            ind_obj.reset_internals()
            if self.verbose:
                log.info(
                    f'{self.label} - {ind_obj.get_name()} '
                    f'start {percent_label}')
            # this will throw on errors to help with debugging
            self.last_ind_obj = ind_obj
            ind_obj.handle_subscribed_dataset(
                algo_id=algo_id,
                ticker=ticker,
                dataset=dataset)
            new_report = ind_obj.get_report()
            if self.verbose:
                log.info(
                    f'{self.label} - {ind_obj.get_name()} '
                    f'end {percent_label} '
                    f'report: {ae_consts.ppj(new_report)}')
            self.latest_report.update(new_report)

            is_buy_value = ind_obj.is_buy
            is_sell_value = ind_obj.is_sell

            """"
            v1 indicator type: supported
            binary decision support on buys and sells
            (like an alert threshold that is on or off)

            v2 indicator type: not supported
            support for buy/sell value range
            (like an alert threshold between a lower and upper bound)
            """
            if (hasattr(ind_obj, 'is_buy') and
                    hasattr(ind_obj, 'is_sell')):
                is_buy_value = ind_obj.is_buy
                is_sell_value = ind_obj.is_sell

            if is_buy_value == ae_consts.INDICATOR_BUY:
                self.latest_report['buys'].append({
                    'cell': idx,
                    'name': ind_obj.get_name(),
                    'id': ind_id,
                    'report': new_report})
            elif is_sell_value == ae_consts.INDICATOR_SELL:
                self.latest_report['sells'].append({
                    'cell': idx,
                    'name': ind_obj.get_name(),
                    'id': ind_id,
                    'report': new_report})
        # end of for all indicators

        self.reports.append(self.latest_report)

        # allow derived indicator processors to build custom reports
        return self.get_latest_report(
            algo_id=algo_id,
            ticker=ticker,
            dataset=dataset)
    # end of process

# end of IndicatorProcessor



"""

Dataset Extraction Utilities
============================

Helper for extracting a dataset from Redis or S3 and
load it into a ``pandas.DataFrame``. This was designed
to ignore the source of the dataset (IEX vs Yahoo) and
perform the extract and load operations without
knowledge of the underlying dataset.

Supported environment variables:

::

    # verbose logging in this module
    export DEBUG_EXTRACT=1

    # verbose logging for just Redis operations in this module
    export DEBUG_REDIS_EXTRACT=1

    # verbose logging for just S3 operations in this module
    export DEBUG_S3_EXTRACT=1

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json

"""

import analysis_engine.consts as ae_consts
import analysis_engine.build_df_from_redis as build_df
import analysis_engine.dataset_scrub_utils as scrub_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def perform_extract(
        df_type,
        df_str,
        work_dict,
        dataset_id_key='ticker',
        scrub_mode='sort-by-date'):
    """perform_extract

    Helper for extracting from Redis or S3

    :param df_type: datafeed type enum
    :param ds_str: dataset string name
    :param work_dict: incoming work request dictionary
    :param dataset_id_key: configurable dataset identifier
                           key for tracking scrubbing and
                           debugging errors
    :param scrub_mode: scrubbing mode on extraction for
                       one-off cleanup before analysis
    """
    status = ae_consts.FAILED
    ds_id = work_dict.get(
        dataset_id_key,
        None)
    label = work_dict.get(
        'label',
        'extract')
    s3_bucket = work_dict.get(
        's3_bucket',
        ae_consts.S3_BUCKET)
    s3_key = work_dict.get(
        's3_key',
        ae_consts.S3_KEY)
    redis_key = work_dict.get(
        'redis_key',
        ae_consts.REDIS_KEY)
    s3_enabled = work_dict.get(
        's3_enabled',
        ae_consts.ENABLED_S3_UPLOAD)
    s3_access_key = work_dict.get(
        's3_access_key',
        ae_consts.S3_ACCESS_KEY)
    s3_secret_key = work_dict.get(
        's3_secret_key',
        ae_consts.S3_SECRET_KEY)
    s3_region_name = work_dict.get(
        's3_region_name',
        ae_consts.S3_REGION_NAME)
    s3_address = work_dict.get(
        's3_address',
        ae_consts.S3_ADDRESS)
    s3_secure = work_dict.get(
        's3_secure',
        ae_consts.S3_SECURE)
    redis_address = work_dict.get(
        'redis_address',
        ae_consts.REDIS_ADDRESS)
    redis_password = work_dict.get(
        'redis_password',
        ae_consts.REDIS_PASSWORD)
    redis_db = work_dict.get(
        'redis_db',
        ae_consts.REDIS_DB)
    redis_expire = work_dict.get(
        'redis_expire',
        ae_consts.REDIS_EXPIRE)

    log.debug(
        f'{label} - {df_str} - START - ds_id={ds_id} scrub_mode={scrub_mode} '
        f'redis_address={redis_address}@{redis_db} redis_key={redis_key} '
        f's3={s3_enabled} s3_address={s3_address} s3_bucket={s3_bucket} '
        f's3_key={s3_key}')

    if ae_consts.ev('DEBUG_REDIS_EXTRACT', '0') == '1':
        log.info(
            f'{label} - {df_str} - ds_id={ds_id} redis '
            f'pw={redis_password} expire={redis_expire}')

    if ae_consts.ev('DEBUG_S3_EXTRACT', '0') == '1':
        log.info(
            f'{label} - {df_str} - ds_id={ds_id} s3 '
            f'ak={s3_access_key} sk={s3_secret_key} '
            f'region={s3_region_name} secure={s3_secure}')

    extract_res = None
    try:
        extract_res = build_df.build_df_from_redis(
            label=label,
            address=redis_address,
            db=redis_db,
            key=redis_key)
    except Exception as e:
        extract_res = None
        log.error(
            f'{label} - {df_str} - ds_id={ds_id} failed extract from '
            f'redis={redis_address}@{redis_db} key={redis_key} ex={e}')
    # end of try/ex extract from redis

    if not extract_res:
        return status, None

    valid_df = (
        extract_res['status'] == ae_consts.SUCCESS
        and extract_res['rec']['valid_df'])

    if not valid_df:
        if ae_consts.ev('DEBUG_S3_EXTRACT', '0') == '1':
            log.error(
                f'{label} - {df_str} ds_id={ds_id} invalid df '
                f'status={ae_consts.get_status(status=extract_res["status"])} '
                f'extract_res={extract_res}')
        return status, None

    extract_df = extract_res['rec']['data']

    log.debug(
        f'{label} - {df_str} ds_id={ds_id} extract scrub={scrub_mode}')

    scrubbed_df = scrub_utils.extract_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=df_type,
        msg_format='df={} date_str={}',
        ds_id=ds_id,
        df=extract_df)

    status = ae_consts.SUCCESS

    return status, scrubbed_df
# end of perform_extract



"""
Build a dictionary by extracting all required pricing datasets
for the algorithm's indicators out of Redis

This dictionary should be passed to an algorithm's ``handle_data``
method like:

.. code-block:: python

    algo.handle_data(build_dataset_node())
"""

import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.api_requests as api_requests
import analysis_engine.iex.extract_df_from_redis as iex_extract_utils
import analysis_engine.td.extract_df_from_redis as td_extract_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def build_dataset_node(
        ticker,
        datasets,
        date=None,
        service_dict=None,
        log_label=None,
        redis_enabled=True,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        redis_key=None,
        s3_enabled=True,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=False,
        s3_key=None,
        verbose=False):
    """build_dataset_node

    Helper for building a dictionary that of
    cached datasets from redis.

    The datasets should be built from
    off the algorithm's config indicators
    ``uses_data`` fields which if not
    set will default to ``minute`` data

    :param ticker: string ticker
    :param datasets: list of string dataset names
        to extract from redis
    :param date: optional - string datetime formatted
        ``YYYY-MM-DD``
        (default is last trading close date)
    :param service_dict: optional - dictionary for all
        service connectivity to Redis and Minio if not
        set the arguments for all ``s3_*`` and ``redis_*``
        will be used to lookup data in Redis and Minio

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``True``)
    :param redis_address: Redis connection string
        format is ``host:port``
        (default is ``localhost:6379``)
    :param redis_db: Redis db to use
        (default is ``0``)
    :param redis_password: optional - Redis password
        (default is ``None``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)
    :param redis_key: optional - redis key not used
        (default is ``None``)

    :param s3_enabled: bool - toggle for turning on/off
        Minio or AWS S3
        (default is ``True``)
    :param s3_address: Minio S3 connection string address
        format is ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)
    :param s3_key: optional s3 key not used
        (default is ``None``)

    **Debugging**

    :param log_label: optional - log label string
    :param verbose: optional - flag for debugging
        (default to ``False``)
    """

    label = log_label
    if not label:
        label = 'build_bt'

    if not date:
        date = ae_utils.get_last_close_str()

    td_convert_to_datetime = (
        ae_consts.TRADIER_CONVERT_TO_DATETIME)

    date_key = f'{ticker}_{date}'

    base_req = api_requests.get_ds_dict(
        ticker=ticker,
        base_key=date_key,
        ds_id=label,
        service_dict=service_dict)

    if not service_dict:
        base_req['redis_enabled'] = redis_enabled
        base_req['redis_address'] = redis_address
        base_req['redis_password'] = redis_password
        base_req['redis_db'] = redis_db
        base_req['redis_key'] = date_key
        base_req['redis_expire'] = redis_expire
        base_req['s3_enabled'] = s3_enabled
        base_req['s3_bucket'] = s3_bucket
        base_req['s3_address'] = s3_address
        base_req['s3_secure'] = s3_secure
        base_req['s3_region_name'] = s3_region_name
        base_req['s3_access_key'] = s3_access_key
        base_req['s3_secret_key'] = s3_secret_key
        base_req['s3_key'] = date_key

    if verbose:
        log.info(
            f'extracting {date_key}')
        """
        for showing connectivity args in the logs
        log.debug(
            f'bt {date_key} {ae_consts.ppj(base_req)}')
        """

    iex_daily_status = ae_consts.FAILED
    iex_minute_status = ae_consts.FAILED
    iex_quote_status = ae_consts.FAILED
    iex_stats_status = ae_consts.FAILED
    iex_peers_status = ae_consts.FAILED
    iex_news_status = ae_consts.FAILED
    iex_financials_status = ae_consts.FAILED
    iex_earnings_status = ae_consts.FAILED
    iex_dividends_status = ae_consts.FAILED
    iex_company_status = ae_consts.FAILED
    td_calls_status = ae_consts.FAILED
    td_puts_status = ae_consts.FAILED

    iex_daily_df = None
    iex_minute_df = None
    iex_quote_df = None
    iex_stats_df = None
    iex_peers_df = None
    iex_news_df = None
    iex_financials_df = None
    iex_earnings_df = None
    iex_dividends_df = None
    iex_company_df = None
    td_calls_df = None
    td_puts_df = None

    if 'daily' in datasets:
        iex_daily_status, iex_daily_df = \
            iex_extract_utils.extract_daily_dataset(
                base_req)
        if iex_daily_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_daily={ticker}')
    if 'minute' in datasets:
        iex_minute_status, iex_minute_df = \
            iex_extract_utils.extract_minute_dataset(
                base_req)
        if iex_minute_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_minute={ticker}')
    if 'quote' in datasets:
        iex_quote_status, iex_quote_df = \
            iex_extract_utils.extract_quote_dataset(
                base_req)
        if iex_quote_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_quote={ticker}')
    if 'stats' in datasets:
        iex_stats_df, iex_stats_df = \
            iex_extract_utils.extract_stats_dataset(
                base_req)
        if iex_stats_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_stats={ticker}')
    if 'peers' in datasets:
        iex_peers_df, iex_peers_df = \
            iex_extract_utils.extract_peers_dataset(
                base_req)
        if iex_peers_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_peers={ticker}')
    if 'news' in datasets:
        iex_news_status, iex_news_df = \
            iex_extract_utils.extract_news_dataset(
                base_req)
        if iex_news_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_news={ticker}')
    if 'financials' in datasets:
        iex_financials_status, iex_financials_df = \
            iex_extract_utils.extract_financials_dataset(
                base_req)
        if iex_financials_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_financials={ticker}')
    if 'earnings' in datasets:
        iex_earnings_status, iex_earnings_df = \
            iex_extract_utils.extract_earnings_dataset(
                base_req)
        if iex_earnings_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_earnings={ticker}')
    if 'dividends' in datasets:
        iex_dividends_status, iex_dividends_df = \
            iex_extract_utils.extract_dividends_dataset(
                base_req)
        if iex_dividends_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_dividends={ticker}')
    if 'company' in datasets:
        iex_company_status, iex_company_df = \
            iex_extract_utils.extract_company_dataset(
                base_req)
        if iex_company_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract iex_company={ticker}')
    # end of iex extracts

    """
    Yahoo disabled on Jan 2019

    yahoo_news_status = ae_consts.FAILED
    yahoo_options_status = ae_consts.FAILED
    yahoo_pricing_status = ae_consts.FAILED
    yahoo_option_calls_df = None
    yahoo_option_puts_df = None
    yahoo_pricing_df = None
    yahoo_news_df = None

    if 'options' in datasets:
        yahoo_options_status, yahoo_option_calls_df = \
            yahoo_extract_utils.extract_option_calls_dataset(
                base_req)
        yahoo_options_status, yahoo_option_puts_df = \
            yahoo_extract_utils.extract_option_puts_dataset(
                base_req)
        if yahoo_options_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract yahoo_options={ticker}')
    if 'pricing' in datasets:
        yahoo_pricing_status, yahoo_pricing_df = \
            yahoo_extract_utils.extract_pricing_dataset(
                base_req)
        if yahoo_pricing_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract yahoo_pricing={ticker}')
    if 'news' in datasets:
        yahoo_news_status, yahoo_news_df = \
            yahoo_extract_utils.extract_yahoo_news_dataset(
                base_req)
        if yahoo_news_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract yahoo_news={ticker}')
    # end of yahoo extracts
    """

    """
    Tradier Extraction
    Debug by setting:

    base_req['verbose_td'] = True
    """
    if (
            'calls' in datasets or
            'tdcalls' in datasets):
        td_calls_status, td_calls_df = \
            td_extract_utils.extract_option_calls_dataset(
                base_req)
        if td_calls_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract tdcalls={ticker}')
        else:
            if ae_consts.is_df(
                    df=td_calls_df):
                for c in td_convert_to_datetime:
                    if c in td_calls_df:
                        td_calls_df[c] = pd.to_datetime(
                            td_calls_df[c],
                            format=ae_consts.COMMON_TICK_DATE_FORMAT)
                if 'date' in td_calls_df:
                    td_calls_df.sort_values(
                        'date',
                        ascending=True)
        # end of converting dates
    # end of Tradier calls extraction

    if (
            'puts' in datasets or
            'tdputs' in datasets):
        td_puts_status, td_puts_df = \
            td_extract_utils.extract_option_puts_dataset(
                base_req)
        if td_puts_status != ae_consts.SUCCESS:
            if verbose:
                log.warn(f'unable to extract tdputs={ticker}')
        else:
            if ae_consts.is_df(
                    df=td_puts_df):
                for c in td_convert_to_datetime:
                    if c in td_puts_df:
                        td_puts_df[c] = pd.to_datetime(
                            td_puts_df[c],
                            format=ae_consts.COMMON_TICK_DATE_FORMAT)
                if 'date' in td_puts_df:
                    td_puts_df.sort_values(
                        'date',
                        ascending=True)
        # end of converting dates
    # end of Tradier puts extraction

    ticker_data = {
        'daily': iex_daily_df,
        'minute': iex_minute_df,
        'quote': iex_quote_df,
        'stats': iex_stats_df,
        'peers': iex_peers_df,
        'news1': iex_news_df,
        'financials': iex_financials_df,
        'earnings': iex_earnings_df,
        'dividends': iex_dividends_df,
        'company': iex_company_df,
        'tdcalls': td_calls_df,
        'tdputs': td_puts_df,
        'calls': None,  # yahoo - here for legacy
        'news': None,  # yahoo - here for legacy
        'pricing': None,  # yahoo - here for legacy
        'puts': None  # yahoo - here for legacy
    }

    return ticker_data
# end of build_dataset_node



"""
Example tool for to profiling algorithm performance for:

- CPU
- Memory
- Profiler
- Heatmap

The pip includes `vprof for profiling algorithm code
performance <https://github.com/nvdv/vprof>`__

#.  Start vprof in remote mode in a first terminal

    .. note:: This command will start a webapp on port ``3434``

    ::

        vprof -r -p 3434

#.  Start Profiler in a second terminal

    .. note:: This command pushes data to the webapp
        in the other terminal listening on port ``3434``

    ::

        vprof -c cm ./analysis_engine/perf/profile_algo_runner.py
"""

import datetime
import vprof.runner as perf_runner
import analysis_engine.consts as ae_consts
import analysis_engine.algo_runner as algo_runner
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(
    name='profile-algo')


def start():
    """start"""

    back_a_few_days = (
        datetime.datetime.now() - datetime.timedelta(days=3))
    start_date = back_a_few_days.strftime(
        ae_consts.COMMON_DATE_FORMAT)

    ticker = 'SPY'
    s3_bucket = 'perftests'
    s3_key = (
        f'{ticker}_{start_date}')

    algo_config = (
        f'./cfg/default_algo.json')
    history_loc = (
        f's3://{s3_bucket}/{s3_key}')

    log.info(
        f'building {ticker} trade history '
        f'start_date={start_date} '
        f'config={algo_config} '
        f'history_loc={history_loc}')

    runner = algo_runner.AlgoRunner(
        ticker=ticker,
        start_date=start_date,
        history_loc=history_loc,
        algo_config=algo_config,
        verbose_algo=False,
        verbose_processor=False,
        verbose_indicators=False)

    runner.start()
# end of start


if __name__ == '__main__':
    perf_runner.run(start, 'cm', args=(), host='0.0.0.0', port=3434)



"""
Work in progress - screener-driven analysis task

**Supported environment variables**

::

    export DEBUG_RESULTS=1

"""

import os
import celery.task as celery_task
import analysis_engine.consts as ae_consts
import analysis_engine.get_task_results as get_task_results
import analysis_engine.fetch as fetch_utils
import analysis_engine.build_result as build_result
import analysis_engine.finviz.fetch_api as finviz_utils
import analysis_engine.work_tasks.custom_task as custom_task
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


@celery_task(
    bind=True,
    base=custom_task.CustomTask,
    queue='task_screener_analysis')
def task_screener_analysis(
        self,
        work_dict):
    """task_screener_analysis

    :param work_dict: task dictionary
    """

    label = work_dict.get(
        'label',
        'screener')

    log.info(f'{label} - start')

    rec = {}
    res = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    """
    Input - Set up dataset sources to collect
    """

    ticker = work_dict.get(
        'ticker',
        None)
    org_tickers = work_dict.get(
        'tickers',
        None)

    if not ticker and not org_tickers:
        res = build_result.build_result(
            status=ae_consts.ERR,
            err='missing ticker or tickers',
            rec=rec)

    tickers = []
    if not org_tickers:
        if ticker:
            tickers = [
                ticker
            ]
    else:
        for t in org_tickers:
            upper_cased_ticker = str(t).upper()
            if upper_cased_ticker not in tickers:
                tickers.append(upper_cased_ticker)
        # build a unique ticker list
    # end of ensuring tickers is a unique list of
    # upper-cased ticker symbol strings

    # fetch from: 'all', 'iex' or 'yahoo'
    fetch_mode = work_dict.get(
        'fetch_mode',
        os.getenv(
            'FETCH_MODE',
            'iex'))
    iex_datasets = work_dict.get(
        'iex_datasets',
        os.getenv(
            'IEX_DATASETS_DEFAULT',
            ae_consts.IEX_DATASETS_DEFAULT))

    # if defined, these are task functions for
    # calling customiized determine Celery tasks
    determine_sells_callback = work_dict.get(
        'determine_sells',
        None)
    determine_buys_callback = work_dict.get(
        'determine_buys',
        None)

    try:

        log.info(
            f'{label} fetch={fetch_mode} tickers={tickers} '
            f'iex_datasets={iex_datasets} '
            f'sell_task={determine_sells_callback} '
            f'buy_task={determine_buys_callback}')

        """
        Input - Set up required urls for building buckets
        """
        fv_urls = work_dict.get(
            'urls',
            )

        if not fv_urls:
            res = build_result.build_result(
                status=ae_consts.ERR,
                err='missing required urls list of screeners',
                rec=rec)

        # stop if something errored out with the
        # celery helper for turning off celery to debug
        # without an engine running
        if res['err']:
            log.error(
                f'{label} - tickers={tickers} fetch={fetch_mode} '
                f'iex_datasets={iex_datasets} hit validation err={res["err"]}')

            return get_task_results.get_task_results(
                work_dict=work_dict,
                result=res)
        # end of input validation checks

        num_urls = len(fv_urls)
        log.info(f'{label} - running urls={fv_urls}')

        fv_dfs = []
        for uidx, url in enumerate(fv_urls):
            log.info(
                f'{label} - url={uidx}/{num_urls} url={url}')
            fv_res = finviz_utils.fetch_tickers_from_screener(
                url=url)
            if fv_res['status'] == ae_consts.SUCCESS:
                fv_dfs.append(fv_res['rec']['data'])
                for ft_tick in fv_res['rec']['tickers']:
                    upper_ft_ticker = ft_tick.upper()
                    if upper_ft_ticker not in tickers:
                        tickers.append(upper_ft_ticker)
                # end of for all found tickers
            else:
                log.error(f'{label} - failed url={uidx}/{num_urls} url={url}')
            # if success vs log the error
        # end of urls to get pandas.DataFrame and unique tickers

        """
        Find tickers in screens
        """

        num_tickers = len(tickers)

        log.info(
            f'{label} - fetching tickers={num_tickers} from urls={num_urls}')

        """
        pull ticker data
        """

        fetch_recs = fetch_utils.fetch(
            tickers=tickers,
            fetch_mode=fetch_mode,
            iex_datasets=iex_datasets)

        if fetch_recs:
            rec = fetch_recs

            """
            Output - Where is data getting cached and archived?
            (this helps to retroactively evaluate trading performance)
            """

            res = build_result.build_result(
                status=ae_consts.SUCCESS,
                err=None,
                rec=rec)
        else:
            err = (
                f'{label} - tickers={ticker} failed fetch={fetch_mode} '
                f'iex_datasets={iex_datasets}')
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)

        log.info(f'{label} - done')
    except Exception as e:
        err = (
            f'{label} - tickers={tickers} fetch={fetch_mode} hit ex={e} ')
        log.error(err)
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=rec)
    # end of try/ex

    return get_task_results.get_task_results(
        work_dict=work_dict,
        result=res)
# end of task_screener_analysis


def run_screener_analysis(
        work_dict):
    """run_screener_analysis

    Celery wrapper for running without celery

    :param work_dict: task data
    """

    fn_name = 'run_screener_analysis'
    label = f'''{fn_name} - {work_dict.get(
        'label',
        '')}'''

    log.info(f'{label} - start')

    response = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec={})
    task_res = {}

    # allow running without celery
    if ae_consts.is_celery_disabled(
            work_dict=work_dict):
        work_dict['celery_disabled'] = True
        task_res = task_screener_analysis(
            work_dict)
        if task_res:
            response = task_res.get(
                'result',
                task_res)
            if ae_consts.ev('DEBUG_RESULTS', '0') == '1':
                response_details = response
                try:
                    response_details = ae_consts.ppj(response)
                except Exception:
                    response_details = response

                log.info(
                    f'{label} task result={response_details}')
        else:
            log.error(
                f'{label} celery was disabled but the task={response} '
                'did not return anything')
        # end of if response
    else:
        task_res = task_screener_analysis.delay(
            work_dict=work_dict)
        rec = {
            'task_id': task_res
        }
        response = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)
    # if celery enabled

    if response:
        if ae_consts.ev('DEBUG_RESULTS', '0') == '1':
            log.info(
                f'{label} - done '
                f'status={ae_consts.get_status(response["status"])} '
                f'err={response["err"]} rec={response["rec"]}')
        else:
            log.info(
                f'{label} - done '
                f'status={ae_consts.get_status(response["status"])} '
                f'err={response["err"]}')
    else:
        log.info(f'{label} - done no response')
    # end of if/else response

    return response
# end of run_screener_analysis



"""
Plot a deep neural network's history output after training

Please check out this `blog post for more information on
how this works <https://
machinelearningmastery.com/custom-metrics-deep-learning-keras-python/>`__
"""

import datetime
import matplotlib.pyplot as plt
import analysis_engine.consts as ae_consts
import analysis_engine.charts as ae_charts
import analysis_engine.build_result as build_result
import spylunking.log.setup_logging as log_utils
from analysis_engine.send_to_slack import post_plot

log = log_utils.build_colorized_logger(name=__name__)


def plot_dnn_fit_history(
        title,
        df,
        red,
        red_color=None,
        red_label=None,
        blue=None,
        blue_color=None,
        blue_label=None,
        green=None,
        green_color=None,
        green_label=None,
        orange=None,
        orange_color=None,
        orange_label=None,
        xlabel='Training Epochs',
        ylabel='Error Values',
        linestyle='-',
        width=9.0,
        height=9.0,
        date_format='%d\n%b',
        df_filter=None,
        start_date=None,
        footnote_text=None,
        footnote_xpos=0.70,
        footnote_ypos=0.01,
        footnote_color='#888888',
        footnote_fontsize=8,
        scale_y=False,
        show_plot=True,
        dropna_for_all=False,
        verbose=False,
        send_plots_to_slack=False):
    """plot_dnn_fit_history

    Plot a DNN's fit history using `Keras fit history object <https://ker
    as.io/visualization/#training-history-visualization>`__

    :param title: title of the plot
    :param df: dataset which is ``pandas.DataFrame``
    :param red: string - column name to plot in
        ``red_color`` (or default ``ae_consts.PLOT_COLORS[red]``)
        where the column is in the ``df`` and
        accessible with:``df[red]``
    :param red_color: hex color code to plot the data in the
        ``df[red]``  (default is ``ae_consts.PLOT_COLORS['red']``)
    :param red_label: optional - string for the label used
        to identify the ``red`` line in the legend
    :param blue: string - column name to plot in
        ``blue_color`` (or default ``ae_consts.PLOT_COLORS['blue']``)
        where the column is in the ``df`` and
        accessible with:``df[blue]``
    :param blue_color: hex color code to plot the data in the
        ``df[blue]``  (default is ``ae_consts.PLOT_COLORS['blue']``)
    :param blue_label: optional - string for the label used
        to identify the ``blue`` line in the legend
    :param green: string - column name to plot in
        ``green_color`` (or default ``ae_consts.PLOT_COLORS['darkgreen']``)
        where the column is in the ``df`` and
        accessible with:``df[green]``
    :param green_color: hex color code to plot the data in the
        ``df[green]``  (default is ``ae_consts.PLOT_COLORS['darkgreen']``)
    :param green_label: optional - string for the label used
        to identify the ``green`` line in the legend
    :param orange: string - column name to plot in
        ``orange_color`` (or default ``ae_consts.PLOT_COLORS['orange']``)
        where the column is in the ``df`` and
        accessible with:``df[orange]``
    :param orange_color: hex color code to plot the data in the
        ``df[orange]``  (default is ``ae_consts.PLOT_COLORS['orange']``)
    :param orange_label: optional - string for the label used
        to identify the ``orange`` line in the legend
    :param xlabel: x-axis label
    :param ylabel: y-axis label
    :param linestyle: style of the plot line
    :param width: float - width of the image
    :param height: float - height of the image
    :param date_format: string - format for dates
    :param df_filter: optional - initialized ``pandas.DataFrame`` query
        for reducing the ``df`` records before plotting. As an eaxmple
        ``df_filter=(df['close'] > 0.01)`` would find only records in
        the ``df`` with a ``close`` value greater than ``0.01``
    :param start_date: optional - string ``datetime``
        for plotting only from a date formatted as
        ``YYYY-MM-DD HH\\:MM\\:SS``
    :param footnote_text: optional - string footnote text
        (default is ``algotraders <DATE>``)
    :param footnote_xpos: optional - float for footnote position
        on the x-axies
        (default is ``0.75``)
    :param footnote_ypos: optional - float for footnote position
        on the y-axies
        (default is ``0.01``)
    :param footnote_color: optional - string hex color code for
        the footnote text
        (default is ``#888888``)
    :param footnote_fontsize: optional - float footnote font size
        (default is ``8``)
    :param scale_y: optional - bool to scale the y-axis with
        .. code-block:: python

            use_ax.set_ylim(
                [0, use_ax.get_ylim()[1] * 3])
    :param show_plot: bool to show the plot
    :param dropna_for_all: optional - bool to toggle keep None's in
        the plot ``df`` (default is drop them for display purposes)
    :param verbose: optional - bool to show logs for debugging
        a dataset
    :param send_plots_to_slack: optional - bool to send the dnn plot to slack
    """

    rec = {
        'ax1': None,
        'fig': None
    }
    result = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    if verbose:
        log.info(
            f'plot_dnn_fit_history - start')

    use_red = red_color
    use_blue = blue_color
    use_green = green_color
    use_orange = orange_color

    if not use_red:
        use_red = ae_consts.PLOT_COLORS['red']
    if not use_blue:
        use_blue = ae_consts.PLOT_COLORS['blue']
    if not use_green:
        use_green = ae_consts.PLOT_COLORS['darkgreen']
    if not use_orange:
        use_orange = ae_consts.PLOT_COLORS['orange']

    use_footnote = footnote_text
    if not use_footnote:
        use_footnote = f'''algotraders - {datetime.datetime.now().strftime(
            ae_consts.COMMON_TICK_DATE_FORMAT)}'''

    column_list = []
    all_plots = []
    if red:
        column_list.append(red)
        all_plots.append({
            'column': red,
            'color': use_red})
    if blue:
        column_list.append(blue)
        all_plots.append({
            'column': blue,
            'color': use_blue})
    if green:
        column_list.append(green)
        all_plots.append({
            'column': green,
            'color': use_green})
    if orange:
        column_list.append(orange)
        all_plots.append({
            'column': orange,
            'color': use_orange})

    use_df = df

    if hasattr(df_filter, 'to_json'):
        # Was seeing this warning below in Jupyter:
        # UserWarning: Boolean Series key
        # will be reindexed to match DataFrame index
        # use_df = use_df[df_filter][column_list]
        # now using:
        use_df = use_df.loc[df_filter, column_list]

    if verbose:
        log.info(
            f'plot_dnn_fit_history '
            f'filter df.index={len(use_df.index)} '
            f'column_list={column_list}')

    ae_charts.set_common_seaborn_fonts()

    hex_color = ae_consts.PLOT_COLORS['blue']
    fig, ax = plt.subplots(
        sharex=True,
        sharey=True,
        figsize=(
            width,
            height))

    all_axes = [
        ax
    ]
    num_plots = len(all_plots)
    for idx, node in enumerate(all_plots):
        column_name = node['column']
        hex_color = node['color']

        use_ax = ax

        if verbose:
            log.info(
                f'plot_dnn_fit_history - '
                f'{idx + 1}/{num_plots} - '
                f'{column_name} '
                f'in '
                f'{hex_color} - '
                f'ax={use_ax}')

        use_ax.plot(
            use_df[column_name],
            label=column_name,
            linestyle=linestyle,
            color=hex_color)
        # end if this is not the fist axis

    # end of for all plots

    lines = []
    for idx, cur_ax in enumerate(all_axes):
        ax_lines = cur_ax.get_lines()
        for line in ax_lines:
            label_name = str(line.get_label())
            use_label = label_name
            if idx == 0:
                if red_label:
                    use_label = red_label
            elif idx == 1:
                if blue_label:
                    use_label = blue_label
            elif idx == 2:
                use_label = label_name[-20:]
                if green_label:
                    use_label = green_label
            elif idx == 3:
                use_label = label_name[-20:]
                if orange_label:
                    use_label = orange_label
            else:
                if len(label_name) > 10:
                    use_label = label_name[-20:]
            # end of fixing the labels in the legend
            line.set_label(use_label)
            if line.get_label() not in lines:
                lines.append(line)
        rec[f'ax{idx + 1}'] = cur_ax
    # end of compiling a new-shortened legend while removing dupes

    for idx, cur_ax in enumerate(all_axes):
        if cur_ax:
            if cur_ax.get_legend():
                cur_ax.get_legend().remove()
    # end of removing all previous legends

    if verbose:
        log.info(
            f'legend lines={[l.get_label() for l in lines]}')
    # log what's going to be in the legend

    ax.legend(
        lines,
        [l.get_label() for l in lines],
        loc='best',
        shadow=True)

    fig.autofmt_xdate()

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    ax.set_title(title)
    ae_charts.add_footnote(
        fig=fig,
        xpos=footnote_xpos,
        ypos=footnote_ypos,
        text=use_footnote,
        color=footnote_color,
        fontsize=footnote_fontsize)
    plt.tight_layout()

    if send_plots_to_slack:
        post_plot(plt, title=title)

    if show_plot:
        plt.show()
    else:
        plt.plot()

    rec['fig'] = fig

    result = build_result.build_result(
        status=ae_consts.SUCCESS,
        err=None,
        rec=rec)

    return result
# end of plot_dnn_fit_history



"""
Dataset Fetch API
"""

import os
import json
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.iex.consts as iex_consts
import analysis_engine.work_tasks.get_new_pricing_data as price_utils
import analysis_engine.iex.extract_df_from_redis as iex_extract_utils
import analysis_engine.yahoo.extract_df_from_redis as yahoo_extract_utils
import analysis_engine.td.extract_df_from_redis as td_extract_utils
import analysis_engine.api_requests as api_requests
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def fetch(
        ticker=None,
        tickers=None,
        fetch_mode=None,
        iex_datasets=None,
        redis_enabled=True,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        s3_enabled=True,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=False,
        celery_disabled=True,
        broker_url=None,
        result_backend=None,
        label=None,
        verbose=False):
    """fetch

    Fetch all supported datasets for a stock ``ticker`` or
    a list of ``tickers`` and returns a dictionary. Once
    run, the datasets will all be cached in Redis and archived
    in Minio (S3) by default.

    Python example:

    .. code-block:: python

        from analysis_engine.fetch import fetch
        d = fetch(ticker='NFLX')
        print(d)
        for k in d['NFLX']:
            print(f'dataset key: {k}')

    By default, it synchronously automates:

        - fetching all datasets
        - caching all datasets in Redis
        - archiving all datasets in Minio (S3)
        - returns all datasets in a single dictionary

    This was created for reducing the amount of typying in
    Jupyter notebooks. It can be set up for use with a
    distributed engine as well with the optional arguments
    depending on your connectitivty requirements.

    .. note:: Please ensure Redis and Minio are running
              before trying to extract tickers

    **Stock tickers to fetch**

    :param ticker: single stock ticker/symbol/ETF to fetch
    :param tickers: optional - list of tickers to fetch

    **(Optional) Data sources, datafeeds and datasets to gather**

    :param fetch_mode: data sources - default is ``all`` (both IEX
        and Yahoo), ``iex`` for only IEX, ``yahoo`` for only Yahoo.
    :param iex_datasets: list of strings for gathering specific `IEX
        datasets <https://iexcloud.io/>`__
        which are set as consts: ``analysis_engine.iex.consts.FETCH_*``.

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``True``)
    :param redis_address: Redis connection string format: ``host:port``
        (default is ``localhost:6379``)
    :param redis_db: Redis db to use
        (default is ``0``)
    :param redis_password: optional - Redis password
        (default is ``None``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)

    **(Optional) Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``True``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)

    **(Optional) Celery worker broker connectivity arguments**

    :param celery_disabled: bool - toggle synchronous mode or publish
        to an engine connected to the `Celery broker and backend
        <https://github.com/celery/celery#transports-and-backends>`__
        (default is ``True`` - synchronous mode without an engine
        or need for a broker or backend for Celery)
    :param broker_url: Celery broker url
        (default is ``redis://0.0.0.0:6379/13``)
    :param result_backend: Celery backend url
        (default is ``redis://0.0.0.0:6379/14``)
    :param label: tracking log label

    **(Optional) Debugging**

    :param verbose: bool - show fetch warnings
        and other debug logging (default is False)

    **Supported environment variables**

    ::

        export REDIS_ADDRESS="localhost:6379"
        export REDIS_DB="0"
        export S3_ADDRESS="localhost:9000"
        export S3_BUCKET="dev"
        export AWS_ACCESS_KEY_ID="trexaccesskey"
        export AWS_SECRET_ACCESS_KEY="trex123321"
        export AWS_DEFAULT_REGION="us-east-1"
        export S3_SECURE="0"
        export WORKER_BROKER_URL="redis://0.0.0.0:6379/13"
        export WORKER_BACKEND_URL="redis://0.0.0.0:6379/14"
    """

    rec = {}

    extract_records = []

    use_tickers = tickers
    if ticker:
        use_tickers = [ticker]
    else:
        if not use_tickers:
            use_tickers = []

    default_iex_datasets = [
        'daily',
        'minute',
        'quote',
        'stats',
        'peers',
        'news',
        'financials',
        'earnings',
        'dividends',
        'company'
    ]

    use_iex_datasets = iex_consts.FETCH_DATASETS
    if len(use_iex_datasets) == 0:
        use_iex_datasets = default_iex_datasets
    if not iex_datasets:
        iex_datasets = use_iex_datasets
    if not fetch_mode:
        fetch_mode = 'all'

    if redis_enabled:
        if not redis_address:
            redis_address = os.getenv(
                'REDIS_ADDRESS',
                'localhost:6379')
        if not redis_password:
            redis_password = os.getenv(
                'REDIS_PASSWORD',
                None)
        if not redis_db:
            redis_db = int(os.getenv(
                'REDIS_DB',
                '0'))
        if not redis_expire:
            redis_expire = os.getenv(
                'REDIS_EXPIRE',
                None)
    if s3_enabled:
        if not s3_address:
            s3_address = os.getenv(
                'S3_ADDRESS',
                'localhost:9000')
        if not s3_access_key:
            s3_access_key = os.getenv(
                'AWS_ACCESS_KEY_ID',
                'trexaccesskey')
        if not s3_secret_key:
            s3_secret_key = os.getenv(
                'AWS_SECRET_ACCESS_KEY',
                'trex123321')
        if not s3_region_name:
            s3_region_name = os.getenv(
                'AWS_DEFAULT_REGION',
                'us-east-1')
        if not s3_secure:
            s3_secure = os.getenv(
                'S3_SECURE',
                '0') == '1'
        if not s3_bucket:
            s3_bucket = os.getenv(
                'S3_BUCKET',
                'dev')
    if not broker_url:
        broker_url = os.getenv(
            'WORKER_BROKER_URL',
            'redis://0.0.0.0:6379/13')
    if not result_backend:
        result_backend = os.getenv(
            'WORKER_BACKEND_URL',
            'redis://0.0.0.0:6379/14')

    if not label:
        label = 'get-latest'

    num_tickers = len(use_tickers)
    last_close_str = ae_utils.get_last_close_str()

    if iex_datasets:
        log.info(
            f'{label} - getting latest for tickers={num_tickers} '
            f'iex={json.dumps(iex_datasets)}')
    else:
        log.info(f'{label} - getting latest for tickers={num_tickers}')

    for ticker in use_tickers:

        ticker_key = f'{ticker}_{last_close_str}'

        fetch_req = api_requests.build_get_new_pricing_request()
        fetch_req['base_key'] = ticker_key
        fetch_req['celery_disabled'] = celery_disabled
        fetch_req['ticker'] = ticker
        fetch_req['label'] = label
        fetch_req['fetch_mode'] = fetch_mode
        fetch_req['iex_datasets'] = iex_datasets
        fetch_req['s3_enabled'] = s3_enabled
        fetch_req['s3_bucket'] = s3_bucket
        fetch_req['s3_address'] = s3_address
        fetch_req['s3_secure'] = s3_secure
        fetch_req['s3_region_name'] = s3_region_name
        fetch_req['s3_access_key'] = s3_access_key
        fetch_req['s3_secret_key'] = s3_secret_key
        fetch_req['s3_key'] = ticker_key
        fetch_req['redis_enabled'] = redis_enabled
        fetch_req['redis_address'] = redis_address
        fetch_req['redis_password'] = redis_password
        fetch_req['redis_db'] = redis_db
        fetch_req['redis_key'] = ticker_key
        fetch_req['redis_expire'] = redis_expire

        fetch_req['redis_address'] = redis_address
        fetch_req['s3_address'] = s3_address

        log.info(
            f'{label} - fetching ticker={ticker} last_close={last_close_str} '
            f'redis_address={fetch_req["redis_address"]} '
            f's3_address={fetch_req["s3_address"]}')

        fetch_res = price_utils.run_get_new_pricing_data(
            work_dict=fetch_req)
        if fetch_res['status'] == ae_consts.SUCCESS:
            log.info(
                f'{label} - fetched ticker={ticker} '
                'preparing for extraction')
            extract_req = fetch_req
            extract_records.append(extract_req)
        else:
            log.warning(
                f'{label} - failed getting ticker={ticker} data '
                f'status={ae_consts.get_status(status=fetch_res["status"])} '
                f'err={fetch_res["err"]}')
        # end of if worked or not
    # end for all tickers to fetch

    """
    Extract Datasets
    """

    iex_daily_status = ae_consts.FAILED
    iex_minute_status = ae_consts.FAILED
    iex_quote_status = ae_consts.FAILED
    iex_stats_status = ae_consts.FAILED
    iex_peers_status = ae_consts.FAILED
    iex_news_status = ae_consts.FAILED
    iex_financials_status = ae_consts.FAILED
    iex_earnings_status = ae_consts.FAILED
    iex_dividends_status = ae_consts.FAILED
    iex_company_status = ae_consts.FAILED
    yahoo_news_status = ae_consts.FAILED
    yahoo_options_status = ae_consts.FAILED
    yahoo_pricing_status = ae_consts.FAILED
    td_calls_status = ae_consts.FAILED
    td_puts_status = ae_consts.FAILED

    iex_daily_df = None
    iex_minute_df = None
    iex_quote_df = None
    iex_stats_df = None
    iex_peers_df = None
    iex_news_df = None
    iex_financials_df = None
    iex_earnings_df = None
    iex_dividends_df = None
    iex_company_df = None
    yahoo_option_calls_df = None
    yahoo_option_puts_df = None
    yahoo_pricing_df = None
    yahoo_news_df = None
    td_calls_df = None
    td_puts_df = None

    extract_iex = True
    if fetch_mode not in ['all', 'iex']:
        extract_iex = False

    extract_yahoo = True
    if fetch_mode not in ['all', 'yahoo']:
        extract_yahoo = False

    extract_td = True
    if fetch_mode not in ['all', 'td']:
        extract_td = False

    for service_dict in extract_records:
        ticker_data = {}
        ticker = service_dict['ticker']

        extract_req = api_requests.get_ds_dict(
            ticker=ticker,
            base_key=service_dict.get('base_key', None),
            ds_id=label,
            service_dict=service_dict)

        if 'daily' in iex_datasets or extract_iex:
            iex_daily_status, iex_daily_df = \
                iex_extract_utils.extract_daily_dataset(
                    extract_req)
            if iex_daily_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_daily={ticker}')
        if 'minute' in iex_datasets or extract_iex:
            iex_minute_status, iex_minute_df = \
                iex_extract_utils.extract_minute_dataset(
                    extract_req)
            if iex_minute_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_minute={ticker}')
        if 'quote' in iex_datasets or extract_iex:
            iex_quote_status, iex_quote_df = \
                iex_extract_utils.extract_quote_dataset(
                    extract_req)
            if iex_quote_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_quote={ticker}')
        if 'stats' in iex_datasets or extract_iex:
            iex_stats_df, iex_stats_df = \
                iex_extract_utils.extract_stats_dataset(
                    extract_req)
            if iex_stats_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_stats={ticker}')
        if 'peers' in iex_datasets or extract_iex:
            iex_peers_df, iex_peers_df = \
                iex_extract_utils.extract_peers_dataset(
                    extract_req)
            if iex_peers_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_peers={ticker}')
        if 'news' in iex_datasets or extract_iex:
            iex_news_status, iex_news_df = \
                iex_extract_utils.extract_news_dataset(
                    extract_req)
            if iex_news_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_news={ticker}')
        if 'financials' in iex_datasets or extract_iex:
            iex_financials_status, iex_financials_df = \
                iex_extract_utils.extract_financials_dataset(
                    extract_req)
            if iex_financials_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_financials={ticker}')
        if 'earnings' in iex_datasets or extract_iex:
            iex_earnings_status, iex_earnings_df = \
                iex_extract_utils.extract_dividends_dataset(
                    extract_req)
            if iex_earnings_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_earnings={ticker}')
        if 'dividends' in iex_datasets or extract_iex:
            iex_dividends_status, iex_dividends_df = \
                iex_extract_utils.extract_dividends_dataset(
                    extract_req)
            if iex_dividends_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_dividends={ticker}')
        if 'company' in iex_datasets or extract_iex:
            iex_company_status, iex_company_df = \
                iex_extract_utils.extract_dividends_dataset(
                    extract_req)
            if iex_company_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch iex_company={ticker}')
        # end of iex extracts

        if extract_yahoo:
            yahoo_options_status, yahoo_option_calls_df = \
                yahoo_extract_utils.extract_option_calls_dataset(
                    extract_req)
            yahoo_options_status, yahoo_option_puts_df = \
                yahoo_extract_utils.extract_option_puts_dataset(
                    extract_req)
            if yahoo_options_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch yahoo_options={ticker}')
            yahoo_pricing_status, yahoo_pricing_df = \
                yahoo_extract_utils.extract_pricing_dataset(
                    extract_req)
            if yahoo_pricing_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch yahoo_pricing={ticker}')
            yahoo_news_status, yahoo_news_df = \
                yahoo_extract_utils.extract_yahoo_news_dataset(
                    extract_req)
            if yahoo_news_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch yahoo_news={ticker}')
        # end of yahoo extracts

        if extract_td:
            td_calls_status, td_calls_df = \
                td_extract_utils.extract_option_calls_dataset(
                    extract_req)
            if td_calls_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch tdcalls={ticker}')
            td_puts_status, td_puts_df = \
                td_extract_utils.extract_option_puts_dataset(
                    extract_req)
            if td_puts_status != ae_consts.SUCCESS:
                if verbose:
                    log.warning(f'unable to fetch tdputs={ticker}')
        # td extracts

        ticker_data['daily'] = iex_daily_df
        ticker_data['minute'] = iex_minute_df
        ticker_data['quote'] = iex_quote_df
        ticker_data['stats'] = iex_stats_df
        ticker_data['peers'] = iex_peers_df
        ticker_data['news1'] = iex_news_df
        ticker_data['financials'] = iex_financials_df
        ticker_data['earnings'] = iex_earnings_df
        ticker_data['dividends'] = iex_dividends_df
        ticker_data['company'] = iex_company_df
        ticker_data['calls'] = yahoo_option_calls_df
        ticker_data['puts'] = yahoo_option_puts_df
        ticker_data['pricing'] = yahoo_pricing_df
        ticker_data['news'] = yahoo_news_df
        ticker_data['tdcalls'] = td_calls_df
        ticker_data['tdputs'] = td_puts_df

        rec[ticker] = ticker_data
    # end of for service_dict in extract_records

    return rec
# end of fetch
